# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import decimal
import typing as t
import uuid
from dataclasses import dataclass, field

from ..._util._warning import _warning
from ._object_annotation import _ObjectAnnotation, annotation_classes
from .num import Num
from .object import Object
from .sensor import Sensor
from .sensor_reference import SensorReference


@dataclass
class Frame:
    """A container of dynamic, timewise, information.

    Parameters
    ----------
    uid: int
        Number of the frame within the annotation file. Must be unique.
    timestamp: decimal.Decimal, optional
        Timestamp containing the Unix epoch time of the frame with up to nanosecond precision.
    sensors: dict of raillabel.format.SensorReference, optional
        References to the sensors with frame specific information like timestamp and uri.
        Default is {}.
    frame_data: dict, optional
        Dictionary containing data directly connected to the frame and not to any object, like
        gps/imu data. Dictionary keys are the ID-strings of the variable the data belongs to.
        Default is {}.
    annotations: dict[str, _ObjectAnnotation subclass], optional
        Dictionary containing all annotations of this frame. Keys are annotation uids.

    Read-Only Attributes
    --------------------
    object_data: dict[str, dict[str, _ObjectAnnotation subclass]]
        Annotations categorized by object. Keys are object uids and values are the annotations
        as a dict, that are part of the object.
    """

    uid: int
    timestamp: t.Optional[decimal.Decimal] = None
    sensors: t.Dict[str, SensorReference] = field(default_factory=dict)
    frame_data: t.Dict[str, Num] = field(default_factory=dict)
    annotations: t.Dict[str, t.Type[_ObjectAnnotation]] = field(default_factory=dict)

    @property
    def object_data(self) -> t.Dict[str, t.Dict[str, t.Type[_ObjectAnnotation]]]:
        """Return annotations categorized by Object-Id.

        Returns
        -------
        dict[str, dict[UUID, _ObjectAnnotation subclass]]
            Dictionary of annotations. Keys are object uids and values are annotations, that are
            contained in the object.
        """

        object_data = {}
        for ann_id, annotation in self.annotations.items():
            if annotation.object.uid not in object_data:
                object_data[annotation.object.uid] = {}

            object_data[annotation.object.uid][ann_id] = annotation

        return object_data

    @classmethod
    def fromdict(
        cls,
        uid: str,
        data_dict: dict,
        objects: t.Dict[str, Object],
        sensors: t.Dict[str, Sensor],
    ) -> "Frame":
        """Generate a Frame object from a dict.

        Parameters
        ----------
        uid: str
            Unique identifier of the frame.
        data_dict: dict
            RailLabel format snippet containing the relevant data.
        objects: dict
            Dictionary of all objects in the scene.
        sensors: dict
            Dictionary of all sensors in the scene.

        Returns
        -------
        frame: raillabel.format.Frame
            Converted Frame object.
        """

        frame = Frame(
            uid=int(uid),
            timestamp=cls._timestamp_fromdict(data_dict),
            sensors=cls._sensors_fromdict(data_dict, int(uid), sensors),
            frame_data=cls._frame_data_fromdict(data_dict, sensors),
            annotations=cls._objects_fromdict(data_dict, int(uid), objects, sensors),
        )

        frame = cls._fix_sensor_uri_attribute(frame)
        return frame

    def asdict(self) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this class instance.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """

        dict_repr = {}

        if self.timestamp is not None or self.sensors != {} or self.frame_data != {}:
            dict_repr["frame_properties"] = {}

        if self.timestamp is not None:
            dict_repr["frame_properties"]["timestamp"] = str(self.timestamp)

        if self.sensors != {}:
            dict_repr["frame_properties"]["streams"] = {
                str(k): v.asdict() for k, v in self.sensors.items()
            }

        if self.frame_data != {}:
            dict_repr["frame_properties"]["frame_data"] = {
                "num": [v.asdict() for v in self.frame_data.values()]
            }

        if self.annotations != {}:
            dict_repr["objects"] = self._annotations_asdict()

        return dict_repr

    @classmethod
    def _timestamp_fromdict(cls, data_dict: dict) -> t.Optional[decimal.Decimal]:

        if "frame_properties" not in data_dict or "timestamp" not in data_dict["frame_properties"]:
            return None

        return decimal.Decimal(data_dict["frame_properties"]["timestamp"])

    @classmethod
    def _sensors_fromdict(
        cls, data_dict: dict, frame_uid: int, scene_sensors: t.Dict[str, Sensor]
    ) -> t.Dict[str, SensorReference]:

        if "frame_properties" not in data_dict or "streams" not in data_dict["frame_properties"]:
            return {}

        sensors = {}

        for sensor_id, sensor_dict in data_dict["frame_properties"]["streams"].items():
            if sensor_id not in scene_sensors:
                _warning(
                    f"{sensor_id} does not exist as a stream, but is referenced in the "
                    + f"sync of frame {frame_uid}."
                )
                continue

            sensors[sensor_id] = SensorReference.fromdict(
                data_dict=sensor_dict, sensor=scene_sensors[sensor_id]
            )

        return sensors

    @classmethod
    def _frame_data_fromdict(
        cls, data_dict: dict, sensors: t.Dict[str, Sensor]
    ) -> t.Dict[str, Num]:

        if "frame_properties" not in data_dict or "frame_data" not in data_dict["frame_properties"]:
            return {}

        frame_data = {}
        for ann_type in data_dict["frame_properties"]["frame_data"]:
            for ann_raw in data_dict["frame_properties"]["frame_data"][ann_type]:
                frame_data[ann_raw["name"]] = Num.fromdict(ann_raw, sensors)

        return frame_data

    @classmethod
    def _objects_fromdict(
        cls,
        data_dict: dict,
        frame_id: int,
        objects: t.Dict[str, Object],
        sensors: t.Dict[str, Sensor],
    ) -> t.Dict[uuid.UUID, t.Type[_ObjectAnnotation]]:

        if "objects" not in data_dict:
            return {}

        annotations = {}

        for obj_id, obj_ann in data_dict["objects"].items():

            if obj_id not in objects:
                _warning(
                    f"{obj_id} does not exist as an object, but is referenced in the object"
                    + f" annotation of frame {frame_id}."
                )
                continue

            object_annotations = cls._object_annotations_fromdict(
                data_dict=obj_ann["object_data"],
                object=objects[obj_id],
                sensors=sensors,
            )

            for annotation in object_annotations:
                if annotation.uid in annotations:
                    cls._issue_duplicate_annotation_uid_warning(annotation.uid, frame_id)
                    annotation.uid = str(uuid.uuid4())

                annotations[annotation.uid] = annotation

        return annotations

    @classmethod
    def _object_annotations_fromdict(
        cls,
        data_dict: dict,
        object: Object,
        sensors: t.Dict[str, Sensor],
    ) -> t.Iterator[t.Type[_ObjectAnnotation]]:

        for ann_type, annotations_raw in data_dict.items():
            for ann_raw in annotations_raw:

                ann_raw = cls._fix_deprecated_annotation_name(ann_raw, ann_type, object.type)

                yield annotation_classes()[ann_type].fromdict(ann_raw, sensors, object)

    @classmethod
    def _fix_sensor_uri_attribute(cls, frame: "Frame") -> "Frame":

        for ann_id, ann in list(frame.annotations.items()):
            for attr_name, attr_val in ann.attributes.items():

                if attr_name != "uri":
                    continue

                _warning(
                    f"Deprecated attribute 'uri' detected in annotation {ann_id}. The error has"
                    + " been fixed. Please update the file via 'raillabel.save()'."
                )

                frame.sensors[ann.sensor.uid].uri = attr_val
                del frame.annotations[ann_id].attributes[attr_name]
                break

        return frame

    @classmethod
    def _fix_deprecated_annotation_name(cls, ann_raw: dict, ann_type: str, obj_type: str) -> dict:

        if "uid" not in ann_raw:
            try:
                ann_raw["uid"] = str(uuid.UUID(ann_raw["name"]))
            except ValueError:
                ann_raw["uid"] = str(uuid.uuid4())

        ann_raw["name"] = f"{ann_raw['coordinate_system']}__{ann_type}__{obj_type}"

        return ann_raw

    @classmethod
    def _issue_duplicate_annotation_uid_warning(cls, ann_uid: str, frame_id: int):
        _warning(
            f"Annotation UID '{ann_uid}' is contained more than once in frame {frame_id}. "
            + "A new uid will be assigned."
        )

    def _annotations_asdict(self) -> dict:
        annotations_dict = {}
        for object_id, annotations in self.object_data.items():
            annotations_dict[object_id] = {"object_data": {}}

            for annotation in annotations.values():
                if annotation.OPENLABEL_ID not in annotations_dict[object_id]["object_data"]:
                    annotations_dict[object_id]["object_data"][annotation.OPENLABEL_ID] = []

                annotations_dict[object_id]["object_data"][annotation.OPENLABEL_ID].append(
                    annotation.asdict()
                )

        return annotations_dict

    def __eq__(self, other) -> bool:
        """Handel equal comparisons."""

        if not hasattr(other, "__dict__"):
            return False

        if len(self.__dict__) != len(other.__dict__):
            return False

        for attr in self.__dict__:

            if type(getattr(self, attr)) == type(self):
                if getattr(self, attr).uid != getattr(other, attr).uid:
                    return False

            else:
                if getattr(self, attr) != getattr(other, attr):
                    return False

        return True



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

import raillabel.format.understand_ai as uai_format

# == Fixtures =========================

@pytest.fixture
def point_3d_uai_dict() -> dict:
    return {
        "x": 0,
        "y": 1,
        "z": 2,
    }

@pytest.fixture
def point_3d_uai() -> dict:
    return uai_format.Point3d(
        x=0,
        y=1,
        z=2,
    )

@pytest.fixture
def point_3d_vec() -> dict:
    return [0, 1, 2]

# == Tests ============================

def test_fromdict():
    point_3d = uai_format.Point3d.fromdict(
        {
            "x": 0,
            "y": 1,
            "z": 2,
        }
    )

    assert point_3d.x == 0
    assert point_3d.y == 1
    assert point_3d.z == 2


if __name__ == "__main__":
    import os
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import sys
from pathlib import Path
from uuid import UUID

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

import raillabel.format.understand_ai as uai_format
from raillabel.format.understand_ai._translation import translate_class_id

# == Fixtures =========================

@pytest.fixture
def bounding_box_2d_uai_dict(sensor_camera_uai_dict, attributes_uai_dict) -> dict:
    return {
        "id": "2f2a1d7f-56d1-435c-a3ec-d6b8fdaaa965",
        "objectId": "48c988bd-76f1-423f-b46d-7e7acb859f31",
        "className": "test_class",
        "geometry": {
            "xMin": 1,
            "yMin": 2,
            "xMax": 3,
            "yMax": 4
        },
        "attributes": attributes_uai_dict,
        "sensor": sensor_camera_uai_dict
    }

@pytest.fixture
def bounding_box_2d_uai(attributes_uai, sensor_camera_uai) -> dict:
    return uai_format.BoundingBox2d(
        id=UUID("2f2a1d7f-56d1-435c-a3ec-d6b8fdaaa965"),
        object_id=UUID("48c988bd-76f1-423f-b46d-7e7acb859f31"),
        class_name="test_class",
        x_min=1,
        y_min=2,
        x_max=3,
        y_max=4,
        attributes=attributes_uai,
        sensor=sensor_camera_uai,
    )

@pytest.fixture
def bounding_box_2d_raillabel_dict(attributes_raillabel_dict, coordinate_system_camera_translated_uid) -> dict:
    return {
        "name": "2f2a1d7f-56d1-435c-a3ec-d6b8fdaaa965",
        "val": [
            2.0,
            3.0,
            2.0,
            2.0
        ],
        "coordinate_system": coordinate_system_camera_translated_uid,
        "attributes": attributes_raillabel_dict,
    }


# == Tests ============================

def test_fromdict(
    attributes_uai_dict, attributes_uai,
    sensor_camera_uai_dict, sensor_camera_uai
):
    bounding_box_2d = uai_format.BoundingBox2d.fromdict(
        {
            "id": "2f2a1d7f-56d1-435c-a3ec-d6b8fdaaa965",
            "objectId": "48c988bd-76f1-423f-b46d-7e7acb859f31",
            "className": "test_class",
            "geometry": {
                "xMin": 1,
                "yMin": 2,
                "xMax": 3,
                "yMax": 4
            },
            "attributes": attributes_uai_dict,
            "sensor": sensor_camera_uai_dict
        }
    )

    assert bounding_box_2d.id == UUID("2f2a1d7f-56d1-435c-a3ec-d6b8fdaaa965")
    assert bounding_box_2d.object_id == UUID("48c988bd-76f1-423f-b46d-7e7acb859f31")
    assert bounding_box_2d.class_name == "test_class"
    assert bounding_box_2d.x_min == 1
    assert bounding_box_2d.y_min == 2
    assert bounding_box_2d.x_max == 3
    assert bounding_box_2d.y_max == 4
    assert bounding_box_2d.attributes == attributes_uai
    assert bounding_box_2d.sensor == sensor_camera_uai


def test_to_raillabel(
    attributes_uai, attributes_raillabel_dict,
    sensor_camera_uai, sensor_camera_raillabel_dict, coordinate_system_camera_translated_uid,
):
    bounding_box_2d = uai_format.BoundingBox2d(
        id=UUID("2f2a1d7f-56d1-435c-a3ec-d6b8fdaaa965"),
        object_id=UUID("48c988bd-76f1-423f-b46d-7e7acb859f31"),
        class_name="test_class",
        x_min=1,
        y_min=2,
        x_max=3,
        y_max=4,
        attributes=attributes_uai,
        sensor=sensor_camera_uai,
    )

    data_dict, object_id, translated_class_id, sensor_reference = bounding_box_2d.to_raillabel()

    assert data_dict == {
        "name": "2f2a1d7f-56d1-435c-a3ec-d6b8fdaaa965",
        "val": [
            2.0,
            3.0,
            2.0,
            2.0
        ],
        "coordinate_system": coordinate_system_camera_translated_uid,
        "attributes": attributes_raillabel_dict,
    }
    assert object_id == str(bounding_box_2d.object_id)
    assert translated_class_id == translate_class_id(bounding_box_2d.class_name)
    assert sensor_reference == sensor_camera_raillabel_dict

if __name__ == "__main__":
    import os
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import decimal
import typing as t
from dataclasses import dataclass

from ..._util._warning import _warning
from .sensor import Sensor


@dataclass
class SensorReference:
    """A reference to a sensor in a specific frame.

    Parameters
    ----------
    sensor: raillabel.format.Sensor
        The sensor this SensorReference corresponds to.
    timestamp: decimal.Decimal
        Timestamp containing the Unix epoch time of the sensor in a specific frame with up to
        nanosecond precision.
    uri: str, optional
        URI to the file corresponding to the frame recording in the particular frame. Default is
        None.
    """

    sensor: Sensor
    timestamp: decimal.Decimal
    uri: t.Optional[str] = None

    @classmethod
    def fromdict(cls, data_dict: dict, sensor: Sensor) -> "SensorReference":
        """Generate a SensorReference object from a dict.

        Parameters
        ----------
        data_dict: dict
            RailLabel format snippet containing the relevant data.
        sensor: raillabel.format.Sensor
            Sensor corresponding to this SensorReference.

        Returns
        -------
        sensor_reference: raillabel.format.SensorReference
            Converted SensorReference object.
        """

        return SensorReference(
            sensor=sensor,
            timestamp=cls._timestamp_fromdict(data_dict["stream_properties"]),
            uri=data_dict.get("uri"),
        )

    def asdict(self) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this class instance.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """

        dict_repr = {"stream_properties": {"sync": {"timestamp": str(self.timestamp)}}}

        if self.uri is not None:
            dict_repr["uri"] = self.uri

        return dict_repr

    @classmethod
    def _timestamp_fromdict(cls, data_dict: dict) -> decimal.Decimal:

        if "stream_sync" in data_dict:
            _warning(
                "Deprecated field 'stream_sync' identified. "
                + "Please update file with raillabel.save()."
            )
            return decimal.Decimal(data_dict["stream_sync"]["timestamp"])

        return decimal.Decimal(data_dict["sync"]["timestamp"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from dataclasses import dataclass
from uuid import UUID

from ._annotation import _Annotation
from .sensor_reference import SensorReference


@dataclass
class BoundingBox2d(_Annotation):
    """A 2d bounding box.

    Parameters
    ----------
    id: uuid.UUID
        Unique identifier of the annotation.
    object_id: uuid.UUID
        Unique identifier of the object this annotation refers to. Used for tracking.
    class_name: str
        Name of the class this annotation belongs to.
    attributes: dict[str, str or list]
        Key value pairs of attributes with the keys beeing the friendly identifier of the
        attribute and the value beeing the attribute value.
    sensor: raillabel.format.understand_ai.SensorReference
        Information about the sensor this annotation is labeled in.
    x_min: float
        Left corner of the bounding box in pixels.
    y_min: float
        Top corner of the bounding box in pixels.
    x_max: float
        Right corner of the bounding box in pixels.
    y_max: float
        Bottom corner of the bounding box in pixels.
    """

    x_min: float
    y_min: float
    x_max: float
    y_max: float

    OPENLABEL_ID = "bbox"

    @classmethod
    def fromdict(cls, data_dict: t.Dict) -> "BoundingBox2d":
        """Generate a BoundingBox2d from a dictionary in the UAI format.

        Parameters
        ----------
        data_dict: dict
            Understand.AI T4 format dictionary containing the data_dict.

        Returns
        -------
        BoundingBox2d
            Converted 2d bounding box.
        """

        return BoundingBox2d(
            id=UUID(data_dict["id"]),
            object_id=UUID(data_dict["objectId"]),
            class_name=data_dict["className"],
            x_min=data_dict["geometry"]["xMin"],
            y_min=data_dict["geometry"]["yMin"],
            x_max=data_dict["geometry"]["xMax"],
            y_max=data_dict["geometry"]["yMax"],
            attributes=data_dict["attributes"],
            sensor=SensorReference.fromdict(data_dict["sensor"]),
        )

    def _val_to_raillabel(self) -> list:
        return [
            (self.x_max + self.x_min) / 2,
            (self.y_max + self.y_min) / 2,
            abs(self.x_max - self.x_min),
            abs(self.y_max - self.y_min),
        ]



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

import raillabel.format.understand_ai as uai_format

# == Fixtures =========================

@pytest.fixture
def size_3d_uai_dict() -> dict:
    return {
        "width": 3,
        "length": 4,
        "height": 5,
    }

@pytest.fixture
def size_3d_uai() -> dict:
    return uai_format.Size3d(
        width=3,
        length=4,
        height=5,
    )

@pytest.fixture
def size_3d_vec() -> dict:
    return [3, 4, 5]

# == Tests ============================

def test_fromdict():
    size_3d = uai_format.Size3d.fromdict(
        {
            "width": 3,
            "length": 4,
            "height": 5,
        }
    )

    assert size_3d.width == 3
    assert size_3d.length == 4
    assert size_3d.height == 5


if __name__ == "__main__":
    import os
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent))

import raillabel


def delete_sensor_from_data(data: dict, sensor_id: str) -> dict:
    del data["openlabel"]["streams"][sensor_id]
    del data["openlabel"]["coordinate_systems"][sensor_id]
    del data["openlabel"]["coordinate_systems"]["base"]["children"][
        data["openlabel"]["coordinate_systems"]["base"]["children"].index(
            sensor_id
        )
    ]

    for frame_id in data["openlabel"]["frames"]:
        if sensor_id not in data["openlabel"]["frames"][frame_id]["frame_properties"]["streams"]:
            continue

        del data["openlabel"]["frames"][frame_id]["frame_properties"]["streams"][sensor_id]

    return data

@pytest.fixture
def loader():
    return raillabel.load_.loader_classes.LoaderRailLabel()


def test_filter_unexpected_kwarg(json_paths):
    # Loads scene
    scene = raillabel.load(json_paths["openlabel_v1_short"], validate=False)

    with pytest.raises(TypeError):
        raillabel.filter(scene, unsupported_kwarg=[])


def test_mutual_exclusivity(json_paths):
    # Loads scene
    scene = raillabel.load(json_paths["openlabel_v1_short"], validate=False)

    with pytest.raises(ValueError):
        raillabel.filter(scene, include_frames=[0], exclude_frames=[1, 2])


def test_filter_frames(json_paths, json_data, loader):
    data = json_data["openlabel_v1_short"]

    # Loads scene
    scene = raillabel.load(json_paths["openlabel_v1_short"], validate=False)

    # Deletes the excluded data
    del data["openlabel"]["frames"]["1"]
    del data["openlabel"]["objects"]["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"]
    data = delete_sensor_from_data(data, "radar")

    # Loads the ground truth filtered data
    scene_filtered_ground_truth = loader.load(data)

    # Tests for include filter
    scene_filtered = raillabel.filter(scene, include_frames=[0])
    assert scene_filtered == scene_filtered_ground_truth

    # Tests for exclude filter
    scene_filtered = raillabel.filter(scene, exclude_frames=[1])
    assert scene_filtered == scene_filtered_ground_truth


def test_filter_start(json_paths, json_data, loader):
    data = json_data["openlabel_v1_short"]

    # Loads scene
    scene = raillabel.load(json_paths["openlabel_v1_short"], validate=False)

    # Deletes the excluded data
    del data["openlabel"]["frames"]["0"]
    del data["openlabel"]["objects"]["b40ba3ad-0327-46ff-9c28-2506cfd6d934"]
    data = delete_sensor_from_data(data, "radar")

    # Loads the ground truth filtered data
    scene_filtered_ground_truth = loader.load(data)

    # Tests for frame filter
    scene_filtered = raillabel.filter(scene, start_frame=1)
    assert scene_filtered == scene_filtered_ground_truth

    # Tests for timestamp filter
    scene_filtered = raillabel.filter(scene, start_timestamp="1632321743.134150")
    assert scene_filtered == scene_filtered_ground_truth


def test_filter_end(json_paths, json_data, loader):
    data = json_data["openlabel_v1_short"]

    # Loads scene
    scene = raillabel.load(json_paths["openlabel_v1_short"], validate=False)

    # Deletes the excluded data
    del data["openlabel"]["frames"]["1"]
    del data["openlabel"]["objects"]["6fe55546-0dd7-4e40-b6b4-bb7ea3445772"]
    data = delete_sensor_from_data(data, "radar")

    # Loads the ground truth filtered data
    scene_filtered_ground_truth = loader.load(data)

    # Tests for frame filter
    scene_filtered = raillabel.filter(scene, end_frame=0)
    assert scene_filtered == scene_filtered_ground_truth

    # Tests for timestamp filter
    scene_filtered = raillabel.filter(scene, end_timestamp="1632321743.233250")
    assert scene_filtered == scene_filtered_ground_truth


def test_filter_object_ids(json_paths, json_data, loader):
    data = json_data["openlabel_v1_short"]

    # Loads scene
    scene = raillabel.load(json_paths["openlabel_v1_short"], validate=False)

    # Deletes the excluded data
    del data["openlabel"]["objects"]["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
    del data["openlabel"]["frames"]["0"]["objects"][
        "22dedd49-6dcb-413b-87ef-00ccfb532e98"
    ]
    del data["openlabel"]["frames"]["1"]["objects"][
        "22dedd49-6dcb-413b-87ef-00ccfb532e98"
    ]
    data = delete_sensor_from_data(data, "radar")

    # Loads the ground truth filtered data
    scene_filtered_ground_truth = loader.load(data)

    # Tests for include filter
    scene_filtered = raillabel.filter(
        scene,
        include_object_ids=[
            "6fe55546-0dd7-4e40-b6b4-bb7ea3445772",
            "b40ba3ad-0327-46ff-9c28-2506cfd6d934",
        ],
    )

    assert scene_filtered == scene_filtered_ground_truth

    # Tests for exclude filter
    scene_filtered = raillabel.filter(
        scene, exclude_object_ids=["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
    )
    assert scene_filtered == scene_filtered_ground_truth


def test_filter_object_types(json_paths, json_data, loader):
    data = json_data["openlabel_v1_short"]

    # Loads scene
    scene = raillabel.load(json_paths["openlabel_v1_short"], validate=False)

    # Deletes the excluded data
    del data["openlabel"]["objects"]["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
    del data["openlabel"]["frames"]["0"]["objects"][
        "22dedd49-6dcb-413b-87ef-00ccfb532e98"
    ]
    del data["openlabel"]["frames"]["1"]["objects"][
        "22dedd49-6dcb-413b-87ef-00ccfb532e98"
    ]
    data = delete_sensor_from_data(data, "radar")

    # Loads the ground truth filtered data
    scene_filtered_ground_truth = loader.load(data)

    # Tests for include filter
    scene_filtered = raillabel.filter(scene, include_object_types=["person"])
    assert scene_filtered == scene_filtered_ground_truth

    # Tests for exclude filter
    scene_filtered = raillabel.filter(scene, exclude_object_types=["train"])
    assert scene_filtered == scene_filtered_ground_truth


def test_filter_annotation_ids(json_paths, json_data, loader):
    data = json_data["openlabel_v1_short"]

    # Loads scene
    scene = raillabel.load(json_paths["openlabel_v1_short"], validate=False)

    # Deletes the excluded data
    del data["openlabel"]["objects"]["22dedd49-6dcb-413b-87ef-00ccfb532e98"]
    del data["openlabel"]["frames"]["0"]["objects"][
        "22dedd49-6dcb-413b-87ef-00ccfb532e98"
    ]
    del data["openlabel"]["frames"]["1"]["objects"][
        "22dedd49-6dcb-413b-87ef-00ccfb532e98"
    ]
    data = delete_sensor_from_data(data, "radar")

    # Loads the ground truth filtered data
    scene_filtered_ground_truth = loader.load(data)

    # Tests for include filter
    scene_filtered = raillabel.filter(
        scene,
        include_annotation_ids=[
            "78f0ad89-2750-4a30-9d66-44c9da73a714",
            "68b4e02c-40c8-4de0-89ad-bc00ed05a043",
            "bebfbae4-61a2-4758-993c-efa846b050a5",
            "3f63201c-fb33-4487-aff6-ae0aa5fa976c",
            "dc2be700-8ee4-45c4-9256-920b5d55c917",
            "c1087f1d-7271-4dee-83ad-519a4e3b78a8",
            "50be7fe3-1f43-47ca-b65a-930e6cfacfeb",
            "6ba42cbc-484e-4b8d-a022-b23c2bb6643c",
            "5f28fa18-8f2a-4a40-a0b6-c0bbedc00f2e",
            "e2503c5d-9fe4-4666-b510-ef644c5a766b",
            "450ceb81-9778-4e63-bf89-42f3ed9f6747",
        ],
    )

    assert scene_filtered == scene_filtered_ground_truth

    # Tests for exclude filter
    scene_filtered = raillabel.filter(
        scene,
        exclude_annotation_ids=[
            "14f58fb0-add7-4ed9-85b3-74615986d854",
            "536ac83a-32c8-4fce-8499-ef32716c64a6",
            "e53bd5e3-980a-4fa7-a0f9-5a2e59ba663c",
            "550df2c3-0e66-483e-bcc6-f3013b7e581b",
            "12b21c52-06ea-4269-9805-e7167e7a74ed",
        ],
    )
    assert scene_filtered == scene_filtered_ground_truth


def test_filter_annotation_types(json_paths, json_data, loader):
    data = json_data["openlabel_v1_short"]

    # Loads scene
    scene = raillabel.load(json_paths["openlabel_v1_short"], validate=False)

    # Deletes the excluded data
    del data["openlabel"]["objects"]["22dedd49-6dcb-413b-87ef-00ccfb532e98"]

    del data["openlabel"]["frames"]["0"]["objects"][
        "b40ba3ad-0327-46ff-9c28-2506cfd6d934"
    ]["object_data"]["cuboid"]
    del data["openlabel"]["frames"]["0"]["objects"][
        "b40ba3ad-0327-46ff-9c28-2506cfd6d934"
    ]["object_data"]["vec"]
    del data["openlabel"]["frames"]["0"]["objects"][
        "22dedd49-6dcb-413b-87ef-00ccfb532e98"
    ]

    del data["openlabel"]["frames"]["1"]["objects"][
        "22dedd49-6dcb-413b-87ef-00ccfb532e98"
    ]
    data = delete_sensor_from_data(data, "radar")

    # Loads the ground truth filtered data
    scene_filtered_ground_truth = loader.load(data)

    # Tests for include filter
    scene_filtered = raillabel.filter(scene, include_annotation_types=["bbox", "poly2d", "Num"])

    assert scene_filtered == scene_filtered_ground_truth

    # Tests for exclude filter
    scene_filtered = raillabel.filter(scene, exclude_annotation_types=["cuboid", "Poly3d", "seg3d"])
    assert scene_filtered == scene_filtered_ground_truth


def test_filter_sensors(json_paths, json_data, loader):
    data = json_data["openlabel_v1_short"]

    # Loads scene
    scene = raillabel.load(json_paths["openlabel_v1_short"], validate=False)

    # Deletes the excluded data
    del data["openlabel"]["objects"]["22dedd49-6dcb-413b-87ef-00ccfb532e98"]

    data = delete_sensor_from_data(data, "lidar")
    data = delete_sensor_from_data(data, "radar")

    del data["openlabel"]["frames"]["0"]["frame_properties"]["frame_data"][
        "num"
    ][-1]
    del data["openlabel"]["frames"]["0"]["objects"][
        "b40ba3ad-0327-46ff-9c28-2506cfd6d934"
    ]["object_data"]["cuboid"]
    del data["openlabel"]["frames"]["0"]["objects"][
        "b40ba3ad-0327-46ff-9c28-2506cfd6d934"
    ]["object_data"]["vec"]
    del data["openlabel"]["frames"]["0"]["objects"][
        "22dedd49-6dcb-413b-87ef-00ccfb532e98"
    ]

    del data["openlabel"]["frames"]["1"]["frame_properties"]["frame_data"][
        "num"
    ][-1]
    del data["openlabel"]["frames"]["1"]["objects"][
        "22dedd49-6dcb-413b-87ef-00ccfb532e98"
    ]

    # Loads the ground truth filtered data
    scene_filtered_ground_truth = loader.load(data)

    # Tests for include filter
    scene_filtered = raillabel.filter(scene, include_sensors=["rgb_middle", "ir_middle"])

    assert scene_filtered == scene_filtered_ground_truth

    # Tests for exclude filter
    scene_filtered = raillabel.filter(scene, exclude_sensors=["lidar"])
    assert scene_filtered == scene_filtered_ground_truth


def test_filter_include_attribute_ids(json_paths, json_data, loader):
    data = json_data["openlabel_v1_short"]

    # Loads scene
    scene = raillabel.load(json_paths["openlabel_v1_short"], validate=False)

    # Deletes the excluded data
    del data["openlabel"]["frames"]["0"]["objects"][
        "22dedd49-6dcb-413b-87ef-00ccfb532e98"
    ]
    del data["openlabel"]["frames"]["0"]["objects"][
        "b40ba3ad-0327-46ff-9c28-2506cfd6d934"
    ]["object_data"]["bbox"][1]
    del data["openlabel"]["frames"]["0"]["objects"][
        "b40ba3ad-0327-46ff-9c28-2506cfd6d934"
    ]["object_data"]["poly2d"][1]
    del data["openlabel"]["frames"]["0"]["objects"][
        "b40ba3ad-0327-46ff-9c28-2506cfd6d934"
    ]["object_data"]["cuboid"][1]
    del data["openlabel"]["frames"]["0"]["objects"][
        "b40ba3ad-0327-46ff-9c28-2506cfd6d934"
    ]["object_data"]["vec"][1]

    del data["openlabel"]["frames"]["1"]["objects"][
        "6fe55546-0dd7-4e40-b6b4-bb7ea3445772"
    ]["object_data"]["bbox"][1]
    del data["openlabel"]["frames"]["1"]["objects"][
        "22dedd49-6dcb-413b-87ef-00ccfb532e98"
    ]["object_data"]["cuboid"][1]
    del data["openlabel"]["frames"]["1"]["objects"][
        "22dedd49-6dcb-413b-87ef-00ccfb532e98"
    ]["object_data"]["vec"][1]

    data = delete_sensor_from_data(data, "ir_middle")
    data = delete_sensor_from_data(data, "radar")

    # Loads the ground truth filtered data
    scene_filtered_ground_truth = loader.load(data)

    # Tests for include filter
    scene_filtered = raillabel.filter(scene, include_attributes={"test_text_attr0": None})

    assert scene_filtered == scene_filtered_ground_truth


def test_filter_exclude_attribute_ids(json_paths, json_data, loader):
    data = json_data["openlabel_v1_short"]

    # Loads scene
    scene = raillabel.load(json_paths["openlabel_v1_short"], validate=False)

    # Deletes the excluded data
    del data["openlabel"]["frames"]["0"]["objects"][
        "b40ba3ad-0327-46ff-9c28-2506cfd6d934"
    ]["object_data"]["bbox"][0]
    del data["openlabel"]["frames"]["0"]["objects"][
        "b40ba3ad-0327-46ff-9c28-2506cfd6d934"
    ]["object_data"]["poly2d"][0]
    del data["openlabel"]["frames"]["0"]["objects"][
        "b40ba3ad-0327-46ff-9c28-2506cfd6d934"
    ]["object_data"]["cuboid"][0]
    del data["openlabel"]["frames"]["0"]["objects"][
        "b40ba3ad-0327-46ff-9c28-2506cfd6d934"
    ]["object_data"]["vec"][0]

    del data["openlabel"]["frames"]["1"]["objects"][
        "6fe55546-0dd7-4e40-b6b4-bb7ea3445772"
    ]["object_data"]["bbox"][0]
    del data["openlabel"]["frames"]["1"]["objects"][
        "6fe55546-0dd7-4e40-b6b4-bb7ea3445772"
    ]["object_data"]["poly2d"][0]
    del data["openlabel"]["frames"]["1"]["objects"][
        "22dedd49-6dcb-413b-87ef-00ccfb532e98"
    ]["object_data"]["cuboid"][0]
    del data["openlabel"]["frames"]["1"]["objects"][
        "22dedd49-6dcb-413b-87ef-00ccfb532e98"
    ]["object_data"]["vec"][0]
    data = delete_sensor_from_data(data, "radar")

    # Loads the ground truth filtered data
    scene_filtered_ground_truth = loader.load(data)

    # Tests for exclude filter
    scene_filtered = raillabel.filter(scene, exclude_attributes={"test_text_attr0": None})
    assert scene_filtered == scene_filtered_ground_truth


def test_filter_exclude_attribute_values(json_paths, json_data, loader):
    data = json_data["openlabel_v1_short"]

    # Loads scene
    scene = raillabel.load(json_paths["openlabel_v1_short"], validate=False)

    # Deletes the excluded data
    del data["openlabel"]["frames"]["0"]["objects"][
        "b40ba3ad-0327-46ff-9c28-2506cfd6d934"
    ]["object_data"]["poly2d"][0]
    data = delete_sensor_from_data(data, "radar")

    # Loads the ground truth filtered data
    scene_filtered_ground_truth = loader.load(data)

    # Tests for exclude filter
    scene_filtered = raillabel.filter(scene, exclude_attributes={"test_num_attr0": 2})
    assert scene_filtered == scene_filtered_ground_truth


# Executes the test if the file is called
if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from decimal import Decimal
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

from raillabel._util._warning import _WarningsLogger
from raillabel.format import Frame

# == Fixtures =========================

@pytest.fixture
def frame_dict(
    sensor_reference_camera_dict,
    num_dict,
    object_person, object_data_person_dict,
    object_train, object_data_train_dict,
) -> dict:
    return {
        "frame_properties": {
            "timestamp": "1632321743.100000072",
            "streams": {
                "rgb_middle": sensor_reference_camera_dict
            },
            "frame_data": {
                "num": [num_dict]
            },
        },
        "objects": {
            object_person.uid: object_data_person_dict,
            object_train.uid: object_data_train_dict,
        }
    }

@pytest.fixture
def frame(
    sensor_reference_camera,
    num,
    all_annotations
) -> dict:
    return Frame(
        uid=0,
        timestamp=Decimal("1632321743.100000072"),
        sensors={sensor_reference_camera.sensor.uid: sensor_reference_camera},
        frame_data={num.name: num},
        annotations=all_annotations
    )

# == Tests ============================

def test_fromdict_sensors(
    sensor_reference_camera_dict,
    sensor_reference_camera,
    sensor_camera
):
    frame = Frame.fromdict(
        uid=0,
        data_dict={
            "frame_properties": {
                "timestamp": "1632321743.100000072",
                "streams": {
                    "rgb_middle": sensor_reference_camera_dict
                }
            }
        },
        sensors={sensor_camera.uid: sensor_camera},
        objects={},
    )

    assert frame.uid == 0
    assert frame.timestamp == Decimal("1632321743.100000072")
    assert frame.sensors == {sensor_reference_camera.sensor.uid: sensor_reference_camera}

def test_fromdict_frame_data(
    num, num_dict,
    sensor_camera
):
    frame = Frame.fromdict(
        uid=1,
        data_dict={
            "frame_properties": {
                "frame_data": {
                    "num": [num_dict]
                }
            }
        },
        sensors={sensor_camera.uid: sensor_camera},
        objects={},
    )

    assert frame.frame_data == {num.name: num}

def test_fromdict_annotations(
    object_data_person_dict, object_person,
    object_data_train_dict, object_train,
    sensors,
    all_annotations,
):
    frame = Frame.fromdict(
        uid=2,
        data_dict={
            "objects": {
                object_person.uid: object_data_person_dict,
                object_train.uid: object_data_train_dict,
            }
        },
        sensors=sensors,
        objects={
            object_person.uid: object_person,
            object_train.uid: object_train,
        },
    )

    assert frame.annotations == all_annotations

def test_fromdict_uri_attribute(
    bbox_dict,
    sensor_reference_camera_dict, sensors,
    object_person, objects,
):
    bbox_with_uri_attribute = bbox_dict
    bbox_with_uri_attribute["attributes"]["text"].append({
        "name": "uri",
        "val": "test_uri.png"
    })

    with _WarningsLogger() as logger:
        frame = Frame.fromdict(
            uid=0,
            data_dict={
                "frame_properties": {
                    "streams": {
                        "rgb_middle": sensor_reference_camera_dict
                    }
                },
                "objects": {
                    object_person.uid: {
                        "object_data": {
                            "bbox": [bbox_with_uri_attribute]
                        }
                    }
                }
            },
            sensors=sensors,
            objects=objects,
        )

    assert len(logger.warnings) == 1
    assert "uri" in logger.warnings[0]
    assert bbox_with_uri_attribute["uid"] in logger.warnings[0]
    assert "raillabel.save()" in logger.warnings[0]

    assert frame.sensors["rgb_middle"].uri == "test_uri.png"
    assert "uri" not in frame.annotations[bbox_with_uri_attribute["uid"]].attributes

def test_fromdict_duplicate_annotation_uid_warning(
    sensors,
    object_person, objects,
    bbox_dict, cuboid_dict
):
    cuboid_dict["uid"] = bbox_dict["uid"]

    with _WarningsLogger() as logger:
        frame = Frame.fromdict(
            uid=2,
            data_dict={
                "objects": {
                    object_person.uid: {
                        "object_data": {
                            "bbox": [bbox_dict],
                            "cuboid": [cuboid_dict],
                        }
                    }
                }
            },
            sensors=sensors,
            objects=objects,
        )

    assert len(logger.warnings) == 1
    assert bbox_dict["uid"] in logger.warnings[0]
    assert list(frame.annotations.values())[0].uid != list(frame.annotations.values())[1].uid


def test_asdict_sensors(
    sensor_reference_camera_dict,
    sensor_reference_camera,
):
    frame = Frame(
        uid=0,
        timestamp=Decimal("1632321743.100000072"),
        sensors={sensor_reference_camera.sensor.uid: sensor_reference_camera},
    )

    assert frame.asdict() == {
        "frame_properties": {
            "timestamp": "1632321743.100000072",
            "streams": {
                "rgb_middle": sensor_reference_camera_dict
            }
        }
    }

def test_asdict_frame_data(num, num_dict):
    frame = Frame(
        uid=0,
        frame_data={num.name: num}
    )

    assert frame.asdict() == {
        "frame_properties": {
            "frame_data": {
                "num": [num_dict]
            }
        }
    }

def test_asdict_object_data(
    object_data_person_dict, object_person,
    object_data_train_dict, object_train,
    all_annotations
):
    frame = Frame(
        uid=0,
        annotations=all_annotations
    )

    assert frame.asdict() == {
        "objects": {
            object_person.uid: object_data_person_dict,
            object_train.uid: object_data_train_dict,
        }
    }


def test_object_data(
    object_person, object_train,
    bbox, cuboid, poly2d, poly3d, seg3d,
    bbox_train
):
    frame = Frame(
        uid=2,
        annotations={
            bbox.uid: bbox,
            poly2d.uid: poly2d,
            cuboid.uid: cuboid,
            poly3d.uid: poly3d,
            seg3d.uid: seg3d,
            bbox_train.uid: bbox_train,
        }
    )

    assert frame.object_data == {
        object_person.uid: {
            bbox.uid: bbox,
            poly2d.uid: poly2d,
            cuboid.uid: cuboid,
            poly3d.uid: poly3d,
            seg3d.uid: seg3d,
        },
        object_train.uid: {
            bbox_train.uid: bbox_train
        }
    }


if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-vv"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
import typing as t
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

from raillabel.format import Sensor, SensorType

# == Fixtures =========================

@pytest.fixture
def sensors(sensor_lidar, sensor_camera, sensor_radar) -> t.Dict[str, Sensor]:
    return {
        sensor_lidar.uid: sensor_lidar,
        sensor_camera.uid: sensor_camera,
        sensor_radar.uid: sensor_radar,
    }

@pytest.fixture
def streams_dict(sensor_camera_dict, sensor_lidar_dict, sensor_radar_dict) -> dict:
    return {
        sensor_camera_dict["uid"]: sensor_camera_dict["stream"],
        sensor_lidar_dict["uid"]: sensor_lidar_dict["stream"],
        sensor_radar_dict["uid"]: sensor_radar_dict["stream"],
    }

@pytest.fixture
def coordinate_systems_dict(sensor_camera_dict, sensor_lidar_dict, sensor_radar_dict) -> dict:
    return {
        "base": {
            "type": "local",
            "parent": "",
            "children": [
                sensor_lidar_dict["uid"],
                sensor_camera_dict["uid"],
                sensor_radar_dict["uid"],
            ]
        },
        sensor_camera_dict["uid"]: sensor_camera_dict["coordinate_system"],
        sensor_lidar_dict["uid"]: sensor_lidar_dict["coordinate_system"],
        sensor_radar_dict["uid"]: sensor_radar_dict["coordinate_system"],
    }


@pytest.fixture
def sensor_lidar_dict(transform_dict) -> dict:
    return {
        "uid": "lidar",
        "stream": {
            "type": "lidar",
            "uri": "/lidar_merged",
        },
        "coordinate_system": {
            "type": "sensor",
            "parent": "base",
            "pose_wrt_parent": transform_dict,
        }
    }

@pytest.fixture
def sensor_lidar(transform) -> Sensor:
    return Sensor(
        uid="lidar",
        extrinsics=transform,
        intrinsics=None,
        type=SensorType.LIDAR,
        uri="/lidar_merged",
    )


@pytest.fixture
def sensor_camera_dict(transform_dict, intrinsics_pinhole_dict) -> dict:
    return {
        "uid": "rgb_middle",
        "stream": {
            "type": "camera",
            "uri": "/S1206063/image",
            "stream_properties": {
                "intrinsics_pinhole": intrinsics_pinhole_dict
            }
        },
        "coordinate_system": {
            "type": "sensor",
            "parent": "base",
            "pose_wrt_parent": transform_dict,
        }
    }

@pytest.fixture
def sensor_camera(transform, intrinsics_pinhole) -> Sensor:
    return Sensor(
        uid="rgb_middle",
        extrinsics=transform,
        intrinsics=intrinsics_pinhole,
        type=SensorType.CAMERA,
        uri="/S1206063/image",
    )


@pytest.fixture
def sensor_radar_dict(transform_dict, intrinsics_radar_dict) -> dict:
    return {
        "uid": "radar",
        "stream": {
            "type": "radar",
            "uri": "/talker1/Nvt/Cartesian",
            "stream_properties": {
                "intrinsics_radar": intrinsics_radar_dict
            }
        },
        "coordinate_system": {
            "type": "sensor",
            "parent": "base",
            "pose_wrt_parent": transform_dict,
        }
    }

@pytest.fixture
def sensor_radar(transform, intrinsics_radar) -> Sensor:
    return Sensor(
        uid="radar",
        extrinsics=transform,
        intrinsics=intrinsics_radar,
        type=SensorType.RADAR,
        uri="/talker1/Nvt/Cartesian",
    )

# == Tests ============================

def test_lidar_fromdict(transform, transform_dict):
    sensor = Sensor.fromdict(
        uid="lidar",
        stream_data_dict={
            "type": "lidar",
            "uri": "/lidar_merged",
        },
        cs_data_dict={
            "type": "sensor",
            "parent": "base",
            "pose_wrt_parent": transform_dict,
        }
    )

    assert sensor.uid == "lidar"
    assert sensor.extrinsics == transform
    assert sensor.intrinsics == None
    assert sensor.type == SensorType.LIDAR
    assert sensor.uri == "/lidar_merged"

def test_lidar_asdict(transform, transform_dict):
    sensor = Sensor(
        uid="lidar",
        extrinsics=transform,
        intrinsics=None,
        type=SensorType.LIDAR,
        uri="/lidar_merged",
    )

    assert sensor.asdict() == {
        "stream": {
            "type": "lidar",
            "uri": "/lidar_merged",
        },
        "coordinate_system": {
            "type": "sensor",
            "parent": "base",
            "pose_wrt_parent": transform_dict,
        }
    }


def test_camera_fromdict(transform, transform_dict, intrinsics_pinhole, intrinsics_pinhole_dict):
    sensor = Sensor.fromdict(
        uid="rgb_middle",
        stream_data_dict={
            "type": "camera",
            "uri": "/S1206063/image",
            "stream_properties": {
                "intrinsics_pinhole": intrinsics_pinhole_dict
            }
        },
        cs_data_dict={
            "type": "sensor",
            "parent": "base",
            "pose_wrt_parent": transform_dict,
        }
    )

    assert sensor.uid == "rgb_middle"
    assert sensor.extrinsics == transform
    assert sensor.intrinsics == intrinsics_pinhole
    assert sensor.type == SensorType.CAMERA
    assert sensor.uri == "/S1206063/image"

def test_camera_asdict(transform, transform_dict, intrinsics_pinhole, intrinsics_pinhole_dict):
    sensor = Sensor(
        uid="rgb_middle",
        extrinsics=transform,
        intrinsics=intrinsics_pinhole,
        type=SensorType.CAMERA,
        uri="/S1206063/image",
    )

    assert sensor.asdict() == {
        "stream": {
            "type": "camera",
            "uri": "/S1206063/image",
            "stream_properties": {
                "intrinsics_pinhole": intrinsics_pinhole_dict
            }
        },
        "coordinate_system": {
            "type": "sensor",
            "parent": "base",
            "pose_wrt_parent": transform_dict,
        }
    }


def test_radar_fromdict(transform, transform_dict, intrinsics_radar, intrinsics_radar_dict):
    sensor = Sensor.fromdict(
        uid="radar",
        stream_data_dict={
            "type": "radar",
            "uri": "/talker1/Nvt/Cartesian",
            "stream_properties": {
                "intrinsics_radar": intrinsics_radar_dict
            }
        },
        cs_data_dict={
            "type": "sensor",
            "parent": "base",
            "pose_wrt_parent": transform_dict,
        }
    )

    assert sensor.uid == "radar"
    assert sensor.extrinsics == transform
    assert sensor.intrinsics == intrinsics_radar
    assert sensor.type == SensorType.RADAR
    assert sensor.uri == "/talker1/Nvt/Cartesian"

def test_radar_asdict(transform, transform_dict, intrinsics_radar, intrinsics_radar_dict):
    sensor = Sensor(
        uid="radar",
        extrinsics=transform,
        intrinsics=intrinsics_radar,
        type=SensorType.RADAR,
        uri="/talker1/Nvt/Cartesian",
    )

    assert sensor.asdict() == {
        "stream": {
            "type": "radar",
            "uri": "/talker1/Nvt/Cartesian",
            "stream_properties": {
                "intrinsics_radar": intrinsics_radar_dict
            }
        },
        "coordinate_system": {
            "type": "sensor",
            "parent": "base",
            "pose_wrt_parent": transform_dict,
        }
    }


if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t

from ._filter_abc import _FilterABC, _ObjectAnnotation


class _FilterObjectTypes(_FilterABC):

    PARAMETERS = ["include_object_types", "exclude_object_types"]
    LEVELS = ["annotation"]

    def passes_filter(self, annotation: t.Type[_ObjectAnnotation]) -> bool:

        if self.include_object_types is not None:
            return annotation.object.type in self.include_object_types

        elif self.exclude_object_types is not None:
            return annotation.object.type not in self.exclude_object_types

        else:
            return True



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

from raillabel.format import Point2d

# == Fixtures =========================

@pytest.fixture
def point2d_dict() -> dict:
    return [1.5, 222]

@pytest.fixture
def point2d() -> dict:
    return Point2d(1.5, 222)


@pytest.fixture
def point2d_another_dict() -> dict:
    return [19, 84]

@pytest.fixture
def point2d_another() -> dict:
    return Point2d(19, 84)

# == Tests ============================

def test_fromdict():
    point2d = Point2d.fromdict(
        [1.5, 222]
    )

    assert point2d.x == 1.5
    assert point2d.y == 222


def test_asdict():
    point2d = Point2d(
        x=1.5,
        y=222,
    )

    assert point2d.asdict() == [1.5, 222]


if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

from raillabel.format import Point3d

# == Fixtures =========================

@pytest.fixture
def point3d_dict() -> dict:
    return [420, 3.14, 0]

@pytest.fixture
def point3d() -> dict:
    return Point3d(420, 3.14, 0)


@pytest.fixture
def point3d_another_dict() -> dict:
    return [9, 8, 7]

@pytest.fixture
def point3d_another() -> dict:
    return Point3d(9, 8, 7)

# == Tests ============================

def test_fromdict():
    point3d = Point3d.fromdict(
        [420, 3.14, 0]
    )

    assert point3d.x == 420
    assert point3d.y == 3.14
    assert point3d.z == 0


def test_asdict():
    point3d = Point3d(
        x=420,
        y=3.14,
        z=0,
    )

    assert point3d.asdict() == [420, 3.14, 0]


if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

import raillabel.format.understand_ai as uai_format

# == Fixtures =========================

@pytest.fixture
def metadata_uai_dict() -> dict:
    return {
        "clip_id": "db_3_2021-09-22-14-28-01_2021-09-22-14-44-03",
        "external_clip_id": "2021-09-22-14-28-01_2021-09-22-14-44-03",
        "project_id": "trains_4",
        "export_time": "2023-04-20 01:38 UTC",
        "exporter_version": "1.0.0",
        "coordinate_system_3d": "FLU",
        "coordinate_system_reference": "SENSOR",
        "folder_name": "2021-09-22-14-28-01_2021-09-22-14-44-03",
    }

@pytest.fixture
def metadata_uai():
    return uai_format.Metadata(
        clip_id="db_3_2021-09-22-14-28-01_2021-09-22-14-44-03",
        external_clip_id="2021-09-22-14-28-01_2021-09-22-14-44-03",
        project_id="trains_4",
        export_time="2023-04-20 01:38 UTC",
        exporter_version="1.0.0",
        coordinate_system_3d="FLU",
        coordinate_system_reference="SENSOR",
        folder_name="2021-09-22-14-28-01_2021-09-22-14-44-03",
    )

@pytest.fixture
def metadata_raillabel_dict(json_data) -> dict:
    return {
        "annotator": "understandAI GmbH",
        "schema_version": "1.0.0",
        "name": "2021-09-22-14-28-01_2021-09-22-14-44-03",
        "subschema_version": json_data["raillabel_schema"]["version"],
        "tagged_file": "2021-09-22-14-28-01_2021-09-22-14-44-03"
    }


# == Tests ============================

def test_fromdict():
    metadata = uai_format.Metadata.fromdict(
        {
            "clip_id": "db_3_2021-09-22-14-28-01_2021-09-22-14-44-03",
            "external_clip_id": "2021-09-22-14-28-01_2021-09-22-14-44-03",
            "project_id": "trains_4",
            "export_time": "2023-04-20 01:38 UTC",
            "exporter_version": "1.0.0",
            "coordinate_system_3d": "FLU",
            "coordinate_system_reference": "SENSOR",
            "folder_name": "2021-09-22-14-28-01_2021-09-22-14-44-03",
        }
    )

    assert metadata.clip_id == "db_3_2021-09-22-14-28-01_2021-09-22-14-44-03"
    assert metadata.external_clip_id == "2021-09-22-14-28-01_2021-09-22-14-44-03"
    assert metadata.project_id == "trains_4"
    assert metadata.export_time == "2023-04-20 01:38 UTC"
    assert metadata.exporter_version == "1.0.0"
    assert metadata.coordinate_system_3d == "FLU"
    assert metadata.coordinate_system_reference == "SENSOR"
    assert metadata.folder_name == "2021-09-22-14-28-01_2021-09-22-14-44-03"


def test_to_raillabel(json_data):
    metadata = uai_format.Metadata(
        clip_id="db_3_2021-09-22-14-28-01_2021-09-22-14-44-03",
        external_clip_id="2021-09-22-14-28-01_2021-09-22-14-44-03",
        project_id="trains_4",
        export_time="2023-04-20 01:38 UTC",
        exporter_version="1.0.0",
        coordinate_system_3d="FLU",
        coordinate_system_reference="SENSOR",
        folder_name="2021-09-22-14-28-01_2021-09-22-14-44-03",
    )

    assert metadata.to_raillabel() == {
        "annotator": "understandAI GmbH",
        "schema_version": "1.0.0",
        "name": "2021-09-22-14-28-01_2021-09-22-14-44-03",
        "subschema_version": json_data["raillabel_schema"]["version"],
        "tagged_file": "2021-09-22-14-28-01_2021-09-22-14-44-03"
    }

if __name__ == "__main__":
    import os
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

import raillabel.format.understand_ai as uai_format

# == Fixtures =========================

@pytest.fixture
def quaternion_uai_dict() -> dict:
    return {
        "x": 0.75318325,
        "y": -0.10270147,
        "z": 0.21430262,
        "w": -0.61338551,
    }

@pytest.fixture
def quaternion_uai() -> dict:
    return uai_format.Quaternion(
        x=0.75318325,
        y=-0.10270147,
        z=0.21430262,
        w=-0.61338551,
    )

@pytest.fixture
def quaternion_vec() -> dict:
    return [0.75318325, -0.10270147,  0.21430262, -0.61338551]

# == Tests ============================

def test_fromdict():
    quaternion = uai_format.Quaternion.fromdict(
        {
            "x": 0.75318325,
            "y": -0.10270147,
            "z": 0.21430262,
            "w": -0.61338551,
        }
    )

    assert quaternion.x == 0.75318325
    assert quaternion.y == -0.10270147
    assert quaternion.z == 0.21430262
    assert quaternion.w == -0.61338551


if __name__ == "__main__":
    import os
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from decimal import Decimal
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

from raillabel.format import SensorReference

# == Fixtures =========================

@pytest.fixture
def sensor_reference_camera_dict() -> dict:
    return {
        "stream_properties": {
            "sync": {
                "timestamp": "1632321743.100000072"
            }
        },
        "uri": "rgb_test0.png"
    }

@pytest.fixture
def sensor_reference_camera(sensor_camera) -> dict:
    return SensorReference(
        sensor=sensor_camera,
        timestamp=Decimal("1632321743.100000072"),
        uri="rgb_test0.png"
    )

# == Tests ============================

def test_fromdict(sensor_camera):
    sensor_reference = SensorReference.fromdict(
        {
            "stream_properties": {
                "sync": {
                    "timestamp": "1632321743.100000072"
                }
            },
            "uri": "rgb_test0.png"
        },
        sensor_camera
    )

    assert sensor_reference.sensor == sensor_camera
    assert sensor_reference.timestamp == Decimal("1632321743.100000072")
    assert sensor_reference.uri == "rgb_test0.png"


def test_asdict(sensor_camera):
    sensor_reference = SensorReference(
        sensor=sensor_camera,
        timestamp=Decimal("1632321743.100000072"),
        uri="rgb_test0.png"
    )

    assert sensor_reference.asdict() == {
        "stream_properties": {
            "sync": {
                "timestamp": "1632321743.100000072"
            }
        },
        "uri": "rgb_test0.png"
    }


if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import sys
from pathlib import Path
from uuid import UUID

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

import raillabel.format.understand_ai as uai_format
from raillabel._util._warning import _WarningsLogger

# == Fixtures =========================

@pytest.fixture
def frame_uai_dict(
    bounding_box_2d_uai_dict,
    bounding_box_3d_uai_dict,
    polygon_2d_uai_dict,
    polyline_2d_uai_dict,
    segmentation_3d_uai_dict,
    sensor_lidar_uai_dict,
) -> dict:
    return {
        "frameId": "000",
        "timestamp": sensor_lidar_uai_dict["timestamp"],
        "annotations": {
            "2D_BOUNDING_BOX": [bounding_box_2d_uai_dict],
            "2D_POLYLINE": [polyline_2d_uai_dict],
            "2D_POLYGON": [polygon_2d_uai_dict],
            "3D_BOUNDING_BOX": [bounding_box_3d_uai_dict],
            "3D_SEGMENTATION": [segmentation_3d_uai_dict],
        }
    }

@pytest.fixture
def frame_uai(
    bounding_box_2d_uai,
    bounding_box_3d_uai,
    polygon_2d_uai,
    polyline_2d_uai,
    segmentation_3d_uai,
    sensor_lidar_uai,
):
    return uai_format.Frame(
        id=0,
        timestamp=sensor_lidar_uai.timestamp,
        bounding_box_2ds={str(bounding_box_2d_uai.id): bounding_box_2d_uai},
        bounding_box_3ds={str(bounding_box_3d_uai.id): bounding_box_3d_uai},
        polygon_2ds={str(polygon_2d_uai.id): polygon_2d_uai},
        polyline_2ds={str(polyline_2d_uai.id): polyline_2d_uai},
        segmentation_3ds={str(segmentation_3d_uai.id): segmentation_3d_uai},
    )

@pytest.fixture
def frame_raillabel_dict(
    bounding_box_2d_uai, bounding_box_2d_raillabel_dict,
    bounding_box_3d_raillabel_dict,
    polygon_2d_uai, polygon_2d_raillabel_dict,
    polyline_2d_uai, polyline_2d_raillabel_dict,
    segmentation_3d_uai, segmentation_3d_raillabel_dict,
    sensor_lidar_uai, sensor_lidar_raillabel_dict, coordinate_system_lidar_translated_uid,
    sensor_camera_raillabel_dict, coordinate_system_camera_translated_uid,
) -> dict:
    return {
        "frame_properties": {
            "timestamp": str(sensor_lidar_uai.timestamp),
            "streams": {
                coordinate_system_camera_translated_uid: sensor_camera_raillabel_dict,
                coordinate_system_lidar_translated_uid: sensor_lidar_raillabel_dict,
            }
        },
        "objects": {
            str(bounding_box_2d_uai.object_id): {
                "object_data": {
                    "bbox": [bounding_box_2d_raillabel_dict],
                    "cuboid": [bounding_box_3d_raillabel_dict],
                }
            },
            str(polygon_2d_uai.object_id): {
                "object_data": {
                    "poly2d": [polygon_2d_raillabel_dict]
                }
            },
            str(polyline_2d_uai.object_id): {
                "object_data": {
                    "poly2d": [polyline_2d_raillabel_dict]
                }
            },
            str(segmentation_3d_uai.object_id): {
                "object_data": {
                    "vec": [segmentation_3d_raillabel_dict]
                }
            },
        }
    }

# == Tests ============================

def test_fromdict(
    bounding_box_2d_uai_dict, bounding_box_2d_uai,
    bounding_box_3d_uai_dict, bounding_box_3d_uai,
    polygon_2d_uai_dict, polygon_2d_uai,
    polyline_2d_uai_dict, polyline_2d_uai,
    segmentation_3d_uai_dict, segmentation_3d_uai,
    sensor_lidar_uai_dict, sensor_lidar_uai,
):
    frame = uai_format.Frame.fromdict(
        {
            "frameId": "000",
            "timestamp": sensor_lidar_uai_dict["timestamp"],
            "annotations": {
                "2D_BOUNDING_BOX": [bounding_box_2d_uai_dict],
                "2D_POLYLINE": [polyline_2d_uai_dict],
                "2D_POLYGON": [polygon_2d_uai_dict],
                "3D_BOUNDING_BOX": [bounding_box_3d_uai_dict],
                "3D_SEGMENTATION": [segmentation_3d_uai_dict],
            }
        }
    )

    assert frame.id == 0
    assert frame.timestamp == sensor_lidar_uai.timestamp
    assert frame.bounding_box_2ds == {str(bounding_box_2d_uai.id): bounding_box_2d_uai}
    assert frame.bounding_box_3ds == {str(bounding_box_3d_uai.id): bounding_box_3d_uai}
    assert frame.polygon_2ds == {str(polygon_2d_uai.id): polygon_2d_uai}
    assert frame.polyline_2ds == {str(polyline_2d_uai.id): polyline_2d_uai}
    assert frame.segmentation_3ds == {str(segmentation_3d_uai.id): segmentation_3d_uai}


def test_to_raillabel(
    bounding_box_2d_uai, bounding_box_2d_raillabel_dict,
    bounding_box_3d_uai, bounding_box_3d_raillabel_dict,
    polygon_2d_uai, polygon_2d_raillabel_dict,
    polyline_2d_uai, polyline_2d_raillabel_dict,
    segmentation_3d_uai, segmentation_3d_raillabel_dict,
    sensor_lidar_uai, sensor_lidar_raillabel_dict, coordinate_system_lidar_translated_uid,
    sensor_camera_raillabel_dict, coordinate_system_camera_translated_uid,
):
    frame = uai_format.Frame(
        id=0,
        timestamp=sensor_lidar_uai.timestamp,
        bounding_box_2ds={str(bounding_box_2d_uai.id): bounding_box_2d_uai},
        bounding_box_3ds={str(bounding_box_3d_uai.id): bounding_box_3d_uai},
        polygon_2ds={str(polygon_2d_uai.id): polygon_2d_uai},
        polyline_2ds={str(polyline_2d_uai.id): polyline_2d_uai},
        segmentation_3ds={str(segmentation_3d_uai.id): segmentation_3d_uai},
    )

    assert frame.to_raillabel() == {
        "frame_properties": {
            "timestamp": str(sensor_lidar_uai.timestamp),
            "streams": {
                coordinate_system_camera_translated_uid: sensor_camera_raillabel_dict,
                coordinate_system_lidar_translated_uid: sensor_lidar_raillabel_dict,
            }
        },
        "objects": {
            str(bounding_box_2d_uai.object_id): {
                "object_data": {
                    "bbox": [bounding_box_2d_raillabel_dict],
                    "cuboid": [bounding_box_3d_raillabel_dict],
                }
            },
            str(polygon_2d_uai.object_id): {
                "object_data": {
                    "poly2d": [polygon_2d_raillabel_dict]
                }
            },
            str(polyline_2d_uai.object_id): {
                "object_data": {
                    "poly2d": [polyline_2d_raillabel_dict]
                }
            },
            str(segmentation_3d_uai.object_id): {
                "object_data": {
                    "vec": [segmentation_3d_raillabel_dict]
                }
            },
        }
    }


def test_warning_duplicate_annotation_id(
    bounding_box_2d_uai_dict, polyline_2d_uai_dict,
    sensor_lidar_uai_dict
):
    polyline_2d_uai_dict["id"] = bounding_box_2d_uai_dict["id"]

    frame_dict = {
        "frameId": "000",
        "timestamp": sensor_lidar_uai_dict["timestamp"],
        "annotations": {
            "2D_BOUNDING_BOX": [bounding_box_2d_uai_dict],
            "2D_POLYLINE": [polyline_2d_uai_dict],
            "2D_POLYGON": [],
            "3D_BOUNDING_BOX": [],
            "3D_SEGMENTATION": [],
        }
    }

    with _WarningsLogger() as logger:
        uai_format.Frame.fromdict(frame_dict)

    assert len(logger.warnings) == 1
    assert bounding_box_2d_uai_dict["id"] in logger.warnings[0]


if __name__ == "__main__":
    import os
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-vv"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from dataclasses import dataclass
from enum import Enum

from .intrinsics_pinhole import IntrinsicsPinhole
from .intrinsics_radar import IntrinsicsRadar
from .point3d import Point3d
from .quaternion import Quaternion
from .transform import Transform


@dataclass
class Sensor:
    """A reference to a physical sensor on the train.

    A sensor in the devkit corresponds to one coordinate_system and one stream in the data format.
    This distinction is set by the OpenLABEL standard, but is not relevant for our data.
    Therefore, we decided to combine these fields.

    Parameters
    ----------
    uid: str
        This is the friendly name of the sensor as well as its identifier. Must be
        unique.
    extrinsics: raillabel.format.Transform, optional
        The external calibration of the sensor defined by the 3D transform to the coordinate
        system origin. Default is None.
    intrinsics: raillabel.format.IntrinsicsPinhole or raillabel.format.IntrinsicsRadar, optional
        The intrinsic calibration of the sensor. Default is None.
    type: raillabel.format.SensorType, optional
        Information about the kind of sensor. Default is None.
    uri: str, optional
        Name of the subdirectory containing the sensor files. Default is None.
    description: str, optional
        Description of the sensor. Default is None.
    """

    uid: str
    extrinsics: t.Optional[Transform] = None
    intrinsics: t.Optional[t.Union[IntrinsicsPinhole, IntrinsicsRadar]] = None
    type: t.Optional["SensorType"] = None
    uri: t.Optional[str] = None
    description: t.Optional[str] = None

    @classmethod
    def fromdict(cls, uid: str, cs_data_dict: dict, stream_data_dict: dict) -> "Sensor":
        """Generate a Sensor object from a dict.

        Parameters
        ----------
        uid: str
            Unique identifier of the sensor.
        cs_data_dict: dict
            RailLabel format dict containing the data about the coordinate system.
        stream_data_dict: dict
            RailLabel format dict containing the data about the stream.

        Returns
        -------
        sensor: raillabel.format.Sensor
            Converted Sensor object.
        """

        return Sensor(
            uid=uid,
            extrinsics=cls._extrinsics_fromdict(cs_data_dict),
            intrinsics=cls._intrinsics_fromdict(
                stream_data_dict, cls._type_fromdict(stream_data_dict)
            ),
            type=cls._type_fromdict(stream_data_dict),
            uri=stream_data_dict.get("uri"),
            description=stream_data_dict.get("description"),
        )

    def asdict(self) -> dict:
        """Export self as a dict compatible with the RailLabel schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this class instance.
        """

        return {
            "coordinate_system": self._as_coordinate_system_dict(),
            "stream": self._as_stream_dict(),
        }

    def _as_coordinate_system_dict(self) -> dict:

        coordinate_system_repr = {"type": "sensor", "parent": "base"}

        if self.extrinsics is not None:
            coordinate_system_repr["pose_wrt_parent"] = self.extrinsics.asdict()

        return coordinate_system_repr

    def _as_stream_dict(self) -> dict:

        stream_repr = {}

        if self.type is not None:
            stream_repr["type"] = str(self.type.value)

        if self.uri is not None:
            stream_repr["uri"] = str(self.uri)

        if self.description is not None:
            stream_repr["description"] = str(self.description)

        if isinstance(self.intrinsics, IntrinsicsPinhole):
            stream_repr["stream_properties"] = {"intrinsics_pinhole": self.intrinsics.asdict()}

        elif isinstance(self.intrinsics, IntrinsicsRadar):
            stream_repr["stream_properties"] = {"intrinsics_radar": self.intrinsics.asdict()}

        return stream_repr

    @classmethod
    def _extrinsics_fromdict(cls, data_dict) -> t.Optional[Transform]:

        if "pose_wrt_parent" not in data_dict:
            return None

        return Transform(
            pos=Point3d(
                x=data_dict["pose_wrt_parent"]["translation"][0],
                y=data_dict["pose_wrt_parent"]["translation"][1],
                z=data_dict["pose_wrt_parent"]["translation"][2],
            ),
            quat=Quaternion(
                x=data_dict["pose_wrt_parent"]["quaternion"][0],
                y=data_dict["pose_wrt_parent"]["quaternion"][1],
                z=data_dict["pose_wrt_parent"]["quaternion"][2],
                w=data_dict["pose_wrt_parent"]["quaternion"][3],
            ),
        )

    @classmethod
    def _intrinsics_fromdict(
        cls, data_dict, sensor_type: t.Optional["SensorType"]
    ) -> t.Optional[IntrinsicsPinhole]:

        if "stream_properties" not in data_dict:
            return None

        if sensor_type == SensorType.CAMERA:

            if "intrinsics_pinhole" in data_dict["stream_properties"]:
                return IntrinsicsPinhole.fromdict(
                    data_dict["stream_properties"]["intrinsics_pinhole"]
                )

        elif sensor_type == SensorType.RADAR:

            if "intrinsics_radar" in data_dict["stream_properties"]:
                return IntrinsicsRadar.fromdict(data_dict["stream_properties"]["intrinsics_radar"])

        return None

    @classmethod
    def _type_fromdict(cls, data_dict) -> t.Optional["SensorType"]:

        if "type" not in data_dict:
            return None

        return SensorType(data_dict["type"])


class SensorType(Enum):
    """Enumeration representing all possible sensor types."""

    CAMERA = "camera"
    LIDAR = "lidar"
    RADAR = "radar"
    GPS_IMU = "gps_imu"
    OTHER = "other"



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import sys
from pathlib import Path
from uuid import UUID

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

import raillabel.format.understand_ai as uai_format
from raillabel.format.understand_ai._translation import translate_class_id

# == Fixtures =========================

@pytest.fixture
def bounding_box_3d_uai_dict(
    point_3d_uai_dict, size_3d_uai_dict, quaternion_uai_dict,
    sensor_lidar_uai_dict,
    attributes_uai_dict
) -> dict:
    return {
        "id": "910399ec-da3e-4d7e-be42-ef8e53e38ca6",
        "objectId": "48c988bd-76f1-423f-b46d-7e7acb859f31",
        "className": "test_class",
        "geometry": {
            "size": size_3d_uai_dict,
            "center": point_3d_uai_dict,
            "quaternion": quaternion_uai_dict
        },
        "attributes": attributes_uai_dict,
        "sensor": sensor_lidar_uai_dict,
    }


@pytest.fixture
def bounding_box_3d_uai(
    point_3d_uai, size_3d_uai, quaternion_uai,
    attributes_uai,
    sensor_lidar_uai
):
    return uai_format.BoundingBox3d(
        id=UUID("910399ec-da3e-4d7e-be42-ef8e53e38ca6"),
        object_id=UUID("48c988bd-76f1-423f-b46d-7e7acb859f31"),
        class_name="test_class",
        size=size_3d_uai,
        center=point_3d_uai,
        quaternion=quaternion_uai,
        attributes=attributes_uai,
        sensor=sensor_lidar_uai,
    )

@pytest.fixture
def bounding_box_3d_raillabel_dict(
    point_3d_vec, size_3d_vec, quaternion_vec,
    coordinate_system_lidar_translated_uid, attributes_raillabel_dict
) -> dict:
    return {
        "name": "910399ec-da3e-4d7e-be42-ef8e53e38ca6",
        "val": point_3d_vec + quaternion_vec + size_3d_vec,
        "coordinate_system": coordinate_system_lidar_translated_uid,
        "attributes": attributes_raillabel_dict
    }


# == Tests ============================

def test_fromdict(
    size_3d_uai_dict, point_3d_uai_dict, quaternion_uai_dict,
    size_3d_uai, point_3d_uai, quaternion_uai,
    sensor_lidar_uai_dict, sensor_lidar_uai,
    attributes_uai_dict, attributes_uai,
):
    bounding_box_3d = uai_format.BoundingBox3d.fromdict(
        {
            "id": "910399ec-da3e-4d7e-be42-ef8e53e38ca6",
            "objectId": "48c988bd-76f1-423f-b46d-7e7acb859f31",
            "className": "test_class",
            "geometry": {
                "size": size_3d_uai_dict,
                "center": point_3d_uai_dict,
                "quaternion": quaternion_uai_dict
            },
            "attributes": attributes_uai_dict,
            "sensor": sensor_lidar_uai_dict,
        }
    )

    assert bounding_box_3d.object_id == UUID("48c988bd-76f1-423f-b46d-7e7acb859f31")
    assert bounding_box_3d.id == UUID("910399ec-da3e-4d7e-be42-ef8e53e38ca6")
    assert bounding_box_3d.class_name == "test_class"
    assert bounding_box_3d.size == size_3d_uai
    assert bounding_box_3d.center == point_3d_uai
    assert bounding_box_3d.quaternion == quaternion_uai
    assert bounding_box_3d.attributes == attributes_uai
    assert bounding_box_3d.sensor == sensor_lidar_uai


def test_to_raillabel(
    size_3d_uai, point_3d_uai, quaternion_uai,
    point_3d_vec, quaternion_vec, size_3d_vec,
    attributes_uai, attributes_raillabel_dict,
    sensor_lidar_uai, coordinate_system_lidar_translated_uid, sensor_lidar_raillabel_dict,
):
    bounding_box_3d = uai_format.BoundingBox3d(
        id=UUID("910399ec-da3e-4d7e-be42-ef8e53e38ca6"),
        object_id=UUID("48c988bd-76f1-423f-b46d-7e7acb859f31"),
        class_name="test_class",
        size=size_3d_uai,
        center=point_3d_uai,
        quaternion=quaternion_uai,
        attributes=attributes_uai,
        sensor=sensor_lidar_uai,
    )

    data_dict, object_id, translated_class_id, sensor_reference = bounding_box_3d.to_raillabel()

    assert data_dict == {
        "name": "910399ec-da3e-4d7e-be42-ef8e53e38ca6",
        "val": point_3d_vec + quaternion_vec + size_3d_vec,
        "coordinate_system": coordinate_system_lidar_translated_uid,
        "attributes": attributes_raillabel_dict
    }
    assert object_id == str(bounding_box_3d.object_id)
    assert translated_class_id == translate_class_id(bounding_box_3d.class_name)
    assert sensor_reference == sensor_lidar_raillabel_dict

if __name__ == "__main__":
    import os
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from abc import ABC, abstractmethod, abstractproperty
from dataclasses import dataclass, field
from importlib import import_module
from inspect import isclass
from pathlib import Path
from pkgutil import iter_modules

from ..._util._attribute_type import AttributeType
from ..._util._warning import _warning
from .object import Object
from .sensor import Sensor


@dataclass
class _ObjectAnnotation(ABC):

    uid: str
    object: Object
    sensor: t.Optional[Sensor] = None
    attributes: t.Dict[str, t.Union[int, float, bool, str, list]] = field(default_factory=dict)

    @property
    def name(self) -> str:
        if self.sensor is None:
            raise AttributeError(
                f"Annotation {self.uid} does not have a 'sensor', which is required "
                + "to create the name."
            )

        return f"{self.sensor.uid}__{self.OPENLABEL_ID}__{self.object.type}"

    @property
    @abstractproperty
    def _REQ_FIELDS(self) -> t.List[str]:
        raise NotImplementedError

    @property
    @abstractproperty
    def OPENLABEL_ID(self) -> t.List[str]:
        raise NotImplementedError

    # === Public Methods =====================================================

    @abstractmethod
    def asdict(self) -> t.Dict:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def fromdict(
        cls,
        data_dict: t.Dict,
        sensors: t.Dict,
        object: Object,
    ) -> t.Type["_ObjectAnnotation"]:
        raise NotImplementedError

    # === Private Methods ====================================================

    def _annotation_required_fields_asdict(self) -> t.Dict:
        """Return the required fields from the parent class to dict."""
        return {
            "uid": str(self.uid),
            "name": str(self.name),
        }

    def _annotation_optional_fields_asdict(self) -> t.Dict:
        """Return the optional fields from the parent class to dict."""

        dict_repr = {}

        if self.sensor is not None:
            dict_repr["coordinate_system"] = str(self.sensor.uid)

        if self.attributes != {}:
            dict_repr["attributes"] = self._attributes_asdict(self.attributes)

        return dict_repr

    def _attributes_asdict(self, attributes: dict) -> dict:
        attributes_dict = {}

        for attr_name, attr_value in attributes.items():

            attr_type = AttributeType.from_value(type(attr_value)).value

            if attr_type not in attributes_dict:
                attributes_dict[attr_type] = []

            attributes_dict[attr_type].append({"name": attr_name, "val": attr_value})

        return attributes_dict

    @classmethod
    def _coordinate_system_fromdict(cls, data_dict: dict, sensors: dict) -> t.Optional[Sensor]:

        is_coordinate_system_in_data = (
            "coordinate_system" in data_dict and data_dict["coordinate_system"] != ""
        )

        if not is_coordinate_system_in_data:
            return None

        if data_dict["coordinate_system"] not in sensors:
            _warning(
                f"'{data_dict['coordinate_system']}' does not exist as a sensor, "
                + f"but is referenced for the annotation {data_dict['uid']}."
            )
            return None

        return sensors[data_dict["coordinate_system"]]

    @classmethod
    def _attributes_fromdict(
        cls,
        data_dict: dict,
    ) -> t.Dict[str, t.Union[int, float, bool, str, list]]:

        if "attributes" not in data_dict:
            return {}

        return {a["name"]: a["val"] for l in data_dict["attributes"].values() for a in l}

    # === Special Methods ====================================================

    def __post_init__(self):
        """Check for required arguments after __init__.

        Inheritance in dataclasses has the flaw, that when the parent class has fields with
        defaults and the child class has fields without, a TypeError is raised due to required
        fields following optional ones. This is solved by making all fields in the child
        optional and checking for required fields in __post_init__().

        Raises
        ------
        TypeError
            If a required field has not been set.
        """

        for f in self._REQ_FIELDS:
            if getattr(self, f) is None:
                raise TypeError(f"{f} is a required argument for {self.__class__.__name__}")


def annotation_classes() -> t.Dict[str, t.Type[_ObjectAnnotation]]:
    """Return dictionary with _Annotation child classes."""
    return ANNOTATION_CLASSES


def _collect_annotation_classes():
    """Collect annotation child classes and store them."""

    global ANNOTATION_CLASSES

    package_dir = str(Path(__file__).resolve().parent)
    for (_, module_name, _) in iter_modules([package_dir]):

        module = import_module(f"raillabel.format.raillabel.{module_name}")
        for attribute_name in dir(module):
            attribute = getattr(module, attribute_name)

            if (
                isclass(attribute)
                and issubclass(attribute, _ObjectAnnotation)
                and attribute != _ObjectAnnotation
            ):
                ANNOTATION_CLASSES[attribute.OPENLABEL_ID] = attribute


ANNOTATION_CLASSES = {}
_collect_annotation_classes()



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

from raillabel.format import Seg3d

# == Fixtures =========================

@pytest.fixture
def seg3d_dict(sensor_lidar, attributes_multiple_types_dict) -> dict:
    return {
        "uid": "db4e4a77-B926-4a6c-a2a6-e0ecf9d8734a",
        "name": "lidar__vec__person",
        "val": [586, 789, 173],
        "coordinate_system": sensor_lidar.uid,
        "attributes": attributes_multiple_types_dict
    }

@pytest.fixture
def seg3d(sensor_lidar, attributes_multiple_types, object_person) -> dict:
    return Seg3d(
        uid="db4e4a77-B926-4a6c-a2a6-e0ecf9d8734a",
        point_ids=[586, 789, 173],
        object=object_person,
        sensor=sensor_lidar,
        attributes=attributes_multiple_types,
    )

# == Tests ============================

def test_fromdict(
    sensor_lidar, sensors,
    object_person,
    attributes_multiple_types, attributes_multiple_types_dict,
):
    seg3d = Seg3d.fromdict(
        {
            "uid": "db4e4a77-B926-4a6c-a2a6-e0ecf9d8734a",
            "name": "lidar__vec__person",
            "val": [586, 789, 173],
            "coordinate_system": sensor_lidar.uid,
            "attributes": attributes_multiple_types_dict
        },
        sensors,
        object_person
    )

    assert seg3d.uid == "db4e4a77-B926-4a6c-a2a6-e0ecf9d8734a"
    assert seg3d.name == "lidar__vec__person"
    assert seg3d.point_ids == [586, 789, 173]
    assert seg3d.object == object_person
    assert seg3d.sensor == sensor_lidar
    assert seg3d.attributes == attributes_multiple_types


def test_asdict(
    sensor_lidar, sensors,
    object_person,
    attributes_multiple_types, attributes_multiple_types_dict,
):
    seg3d = Seg3d(
        uid="db4e4a77-B926-4a6c-a2a6-e0ecf9d8734a",
        point_ids=[586, 789, 173],
        object=object_person,
        sensor=sensor_lidar,
        attributes=attributes_multiple_types,
    )

    assert seg3d.asdict() == {
        "uid": "db4e4a77-B926-4a6c-a2a6-e0ecf9d8734a",
        "name": "lidar__vec__person",
        "val": [586, 789, 173],
        "coordinate_system": sensor_lidar.uid,
        "attributes": attributes_multiple_types_dict
    }


if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
from dataclasses import dataclass
from uuid import UUID

from ._annotation import _Annotation
from .point_3d import Point3d
from .quaternion import Quaternion
from .sensor_reference import SensorReference
from .size_3d import Size3d


@dataclass
class BoundingBox3d(_Annotation):
    """A 3d bounding box.

    Parameters
    ----------
    id: uuid.UUID
        Unique identifier of the annotation.
    object_id: uuid.UUID
        Unique identifier of the object this annotation refers to. Used for tracking.
    class_name: str
        Name of the class this annotation belongs to.
    attributes: dict[str, str or list]
        Key value pairs of attributes with the keys beeing the friendly identifier of the
        attribute and the value beeing the attribute value.
    sensor: raillabel.format.understand_ai.SensorReference
        Information about the sensor this annotation is labeled in.
    center: raillabel.format.understand_ai.Point3d
        Center position of the bounding box.
    size: raillabel.format.understand_ai.Size3d
        3d size of the bounding box.
    quaternion: raillabel.format.understand_ai.Quaternion
        Rotation quaternion of the bounding box.
    """

    center: Point3d
    size: Size3d
    quaternion: Quaternion

    OPENLABEL_ID = "cuboid"

    @classmethod
    def fromdict(cls, data_dict: t.Dict) -> "BoundingBox3d":
        """Generate a BoundingBox3d from a dictionary in the UAI format.

        Parameters
        ----------
        data_dict: dict
            Understand.AI T4 format dictionary containing the data_dict.

        Returns
        -------
        BoundingBox3d
            Converted 3d bounding box.
        """

        return BoundingBox3d(
            id=UUID(data_dict["id"]),
            object_id=UUID(data_dict["objectId"]),
            class_name=data_dict["className"],
            center=Point3d.fromdict(data_dict["geometry"]["center"]),
            size=Size3d.fromdict(data_dict["geometry"]["size"]),
            quaternion=Quaternion.fromdict(data_dict["geometry"]["quaternion"]),
            attributes=data_dict["attributes"],
            sensor=SensorReference.fromdict(data_dict["sensor"]),
        )

    def _val_to_raillabel(self) -> list:
        return [
            float(self.center.x),
            float(self.center.y),
            float(self.center.z),
            float(self.quaternion.x),
            float(self.quaternion.y),
            float(self.quaternion.z),
            float(self.quaternion.w),
            float(self.size.width),
            float(self.size.length),
            float(self.size.height),
        ]



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

import raillabel.format.understand_ai as uai_format
from raillabel._util._warning import _WarningsLogger

# == Fixtures =========================

@pytest.fixture
def scene_uai_dict(
    metadata_uai_dict,
    coordinate_system_camera_uai_dict,
    coordinate_system_lidar_uai_dict,
    coordinate_system_radar_uai_dict,
    frame_uai_dict,
) -> dict:
    return {
        "metadata": metadata_uai_dict,
        "coordinateSystems": [
            coordinate_system_camera_uai_dict,
            coordinate_system_lidar_uai_dict,
            coordinate_system_radar_uai_dict,
        ],
        "frames": [frame_uai_dict]
    }

@pytest.fixture
def scene_uai(
    metadata_uai,
    coordinate_system_camera_uai,
    coordinate_system_lidar_uai,
    coordinate_system_radar_uai,
    frame_uai,
):
    return uai_format.Scene(
        metadata=metadata_uai,
        coordinate_systems={
            coordinate_system_camera_uai.uid: coordinate_system_camera_uai,
            coordinate_system_lidar_uai.uid: coordinate_system_lidar_uai,
            coordinate_system_radar_uai.uid: coordinate_system_radar_uai,
        },
        frames={
            frame_uai.id: frame_uai
        },
    )

@pytest.fixture
def scene_raillabel_dict(
    metadata_raillabel_dict,
    coordinate_system_camera_uai, coordinate_system_camera_raillabel_dict,
    coordinate_system_lidar_uai, coordinate_system_lidar_raillabel_dict,
    coordinate_system_radar_uai, coordinate_system_radar_raillabel_dict,
) -> dict:
    return

# == Tests ============================

def test_fromdict(
    metadata_uai_dict, metadata_uai,
    coordinate_system_camera_uai_dict, coordinate_system_camera_uai,
    coordinate_system_lidar_uai_dict, coordinate_system_lidar_uai,
    coordinate_system_radar_uai_dict, coordinate_system_radar_uai,
    frame_uai_dict, frame_uai,
):
    scene = uai_format.Scene.fromdict(
        {
            "metadata": metadata_uai_dict,
            "coordinateSystems": [
                coordinate_system_camera_uai_dict,
                coordinate_system_lidar_uai_dict,
                coordinate_system_radar_uai_dict,
            ],
            "frames": [frame_uai_dict]
        }
    )

    assert scene.metadata == metadata_uai
    assert scene.coordinate_systems == {
        coordinate_system_camera_uai.uid: coordinate_system_camera_uai,
        coordinate_system_lidar_uai.uid: coordinate_system_lidar_uai,
        coordinate_system_radar_uai.uid: coordinate_system_radar_uai,
    }
    assert scene.frames == {
        frame_uai.id: frame_uai
    }

def test_fromdict_duplicate_frame_id_warning(
    metadata_uai_dict,
    coordinate_system_camera_uai_dict,
    coordinate_system_lidar_uai_dict,
    coordinate_system_radar_uai_dict,
    frame_uai_dict,
):
    with _WarningsLogger() as logger:
        scene = uai_format.Scene.fromdict(
            {
                "metadata": metadata_uai_dict,
                "coordinateSystems": [
                    coordinate_system_camera_uai_dict,
                    coordinate_system_lidar_uai_dict,
                    coordinate_system_radar_uai_dict,
                ],
                "frames": [frame_uai_dict, frame_uai_dict]
            }
        )

    assert len(logger.warnings) == 1
    assert "0" in logger.warnings[0]
    assert len(scene.frames) == 1


def test_to_raillabel__metadata(metadata_uai, metadata_raillabel_dict):
    scene = uai_format.Scene(
        metadata=metadata_uai,
        coordinate_systems={},
        frames={},
    )

    assert scene.to_raillabel()["openlabel"]["metadata"] == metadata_raillabel_dict

def test_to_raillabel__sensors(
    metadata_uai,
    coordinate_system_camera_uai, coordinate_system_lidar_uai,
    coordinate_system_camera_raillabel_dict, coordinate_system_lidar_raillabel_dict,
    coordinate_system_camera_translated_uid, coordinate_system_lidar_translated_uid,
):
    scene = uai_format.Scene(
        metadata=metadata_uai,
        coordinate_systems={
            coordinate_system_camera_uai.uid: coordinate_system_camera_uai,
            coordinate_system_lidar_uai.uid: coordinate_system_lidar_uai,
        },
        frames={},
    )

    assert scene.to_raillabel()["openlabel"]["coordinate_systems"] == {
        "base": {
            "type": "local",
            "parent": "",
            "children": [
                coordinate_system_camera_translated_uid,
                coordinate_system_lidar_translated_uid
            ]
        },
        coordinate_system_camera_translated_uid: coordinate_system_camera_raillabel_dict[0],
        coordinate_system_lidar_translated_uid: coordinate_system_lidar_raillabel_dict[0],
    }
    assert scene.to_raillabel()["openlabel"]["streams"] == {
        coordinate_system_camera_translated_uid: coordinate_system_camera_raillabel_dict[1],
        coordinate_system_lidar_translated_uid: coordinate_system_lidar_raillabel_dict[1],
    }

def test_to_raillabel__frames(
    metadata_uai,
    coordinate_system_camera_uai, coordinate_system_lidar_uai,
    frame_uai, frame_raillabel_dict
):
    scene = uai_format.Scene(
        metadata=metadata_uai,
        coordinate_systems={
            coordinate_system_camera_uai.uid: coordinate_system_camera_uai,
            coordinate_system_lidar_uai.uid: coordinate_system_lidar_uai,
        },
        frames={
            frame_uai.id: frame_uai
        },
    )

    assert scene.to_raillabel()["openlabel"]["frames"] == {
        str(frame_uai.id): frame_raillabel_dict
    }


if __name__ == "__main__":
    import os
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-vv"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

from raillabel.format import Quaternion

# == Fixtures =========================

@pytest.fixture
def quaternion_dict() -> dict:
    return [0.75318325, -0.10270147,  0.21430262, -0.61338551]

@pytest.fixture
def quaternion() -> dict:
    return Quaternion(0.75318325, -0.10270147,  0.21430262, -0.61338551)

# == Tests ============================

def test_fromdict():
    quaternion = Quaternion.fromdict(
        [
            0.75318325,
            -0.10270147,
            0.21430262,
            -0.61338551
        ]
    )

    assert quaternion.x == 0.75318325
    assert quaternion.y == -0.10270147
    assert quaternion.z == 0.21430262
    assert quaternion.w == -0.61338551


def test_asdict():
    quaternion = Quaternion(
        x=0.75318325,
        y=-0.10270147,
        z=0.21430262,
        w=-0.61338551
    )

    assert quaternion.asdict() == [
        0.75318325,
        -0.10270147,
        0.21430262,
        -0.61338551
    ]


if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import random
import sys
import typing as t
from pathlib import Path
from uuid import UUID, uuid4

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

from raillabel._util._attribute_type import AttributeType
from raillabel.format import (
    Bbox,
    Cuboid,
    Frame,
    FrameInterval,
    Object,
    Point2d,
    Point3d,
    Quaternion,
    Sensor,
    Size2d,
    Size3d,
)

# == Fixtures =========================

@pytest.fixture
def objects_dict(object_person_dict, object_train_dict) -> dict:
    return {
        object_person_dict["object_uid"]: object_person_dict["data_dict"],
        object_train_dict["object_uid"]: object_train_dict["data_dict"],
    }

@pytest.fixture
def objects(object_person, object_train) -> t.Dict[str, Object]:
    return {
        object_person.uid: object_person,
        object_train.uid: object_train,
    }


@pytest.fixture
def object_person_dict() -> dict:
    return {
        "object_uid": "b40ba3ad-0327-46ff-9c28-2506cfd6d934",
        "data_dict": {
            "name": "person_0000",
            "type": "person",
        },
    }

@pytest.fixture
def object_person() -> dict:
    return Object(
        uid="b40ba3ad-0327-46ff-9c28-2506cfd6d934",
        name="person_0000",
        type="person",
    )


@pytest.fixture
def object_train_dict() -> dict:
    return {
        "object_uid": "d51a19be-8bc2-4a82-b66a-03c8de95b0cf",
        "data_dict": {
            "name": "train_0000",
            "type": "train",
        },
    }

@pytest.fixture
def object_train() -> dict:
    return Object(
        uid="d51a19be-8bc2-4a82-b66a-03c8de95b0cf",
        name="train_0000",
        type="train",
    )

# == Tests ============================

def test_fromdict():
    object = Object.fromdict(
        object_uid="b40ba3ad-0327-46ff-9c28-2506cfd6d934",
        data_dict={
            "name": "person_0000",
            "type": "person",
        }
    )

    assert object.uid == "b40ba3ad-0327-46ff-9c28-2506cfd6d934"
    assert object.name == "person_0000"
    assert object.type == "person"


def test_asdict_no_frames():
    object = Object(
        uid="b40ba3ad-0327-46ff-9c28-2506cfd6d934",
        name="person_0000",
        type="person",
    )

    assert object.asdict() == {
        "name": "person_0000",
        "type": "person",
    }

def test_asdict_with_frames():
    object = Object(
        uid="b40ba3ad-0327-46ff-9c28-2506cfd6d934",
        name="person_0000",
        type="person",
    )

    frames = {
        0: build_frame(0,
            {
                object: [build_annotation("rgb_middle__bbox__person", object)]
            }
        ),
    }

    object_dict = object.asdict(frames)

    assert "frame_intervals" in object_dict
    assert "object_data_pointers" in object_dict
    assert "rgb_middle__bbox__person" in object_dict["object_data_pointers"]


def test_frame_intervals():
    object = Object(
        uid="b40ba3ad-0327-46ff-9c28-2506cfd6d934",
        name="person_0000",
        type="person",
    )

    frames = {
        0: build_frame(0, {object: [build_annotation("rgb_middle__bbox__person", object)]}),
        1: build_frame(1, {object: [build_annotation("rgb_middle__bbox__person", object)]}),
        2: build_frame(2, {}),
        3: build_frame(3, {object: [build_annotation("rgb_middle__bbox__person", object)]}),
    }

    assert object.frame_intervals(frames) == [
        FrameInterval(0, 1),
        FrameInterval(3, 3),
    ]


def test_object_data_pointers__sensor():
    object = build_object("person")

    frames = {
        0: build_frame(0,
            {
                object: [
                    build_annotation("rgb_middle__bbox__person", object),
                    build_annotation("lidar__bbox__person", object)
                ]
            }
        )
    }

    object_data_pointers = object.object_data_pointers(frames)

    assert set(object_data_pointers.keys()) == set(["rgb_middle__bbox__person", "lidar__bbox__person"])

def test_object_data_pointers__annotation_type():
    object = build_object("person")

    frames = {
        0: build_frame(0,
            {
                object: [
                    build_annotation("rgb_middle__bbox__person", object),
                    build_annotation("rgb_middle__cuboid__person", object)
                ]
            }
        )
    }

    object_data_pointers = object.object_data_pointers(frames)

    assert set(object_data_pointers.keys()) == set([
        "rgb_middle__bbox__person",
        "rgb_middle__cuboid__person"
    ])

def test_object_data_pointers__one_frame_interval():
    object = build_object("person")

    frames = {
        0: build_frame(0,
            {
                object: [build_annotation("rgb_middle__bbox__person", object)]
            }
        ),
        1: build_frame(1,
            {
                object: [build_annotation("rgb_middle__bbox__person", object)]
            }
        ),
    }

    object_data_pointers = object.object_data_pointers(frames)

    assert len(object_data_pointers) == 1
    assert object_data_pointers["rgb_middle__bbox__person"].frame_intervals == [
        FrameInterval(0, 1)
    ]

def test_object_data_pointers__two_frame_intervals():
    object = build_object("person")

    frames = {
        0: build_frame(0,
            {
                object: [build_annotation("rgb_middle__bbox__person", object)]
            }
        ),
        1: build_frame(1,
            {
                object: [build_annotation("rgb_middle__bbox__person", object)]
            }
        ),
        8: build_frame(8,
            {
                object: [build_annotation("rgb_middle__bbox__person", object)]
            }
        ),
    }

    object_data_pointers = object.object_data_pointers(frames)

    assert len(object_data_pointers) == 1
    assert object_data_pointers["rgb_middle__bbox__person"].frame_intervals == [
        FrameInterval(0, 1),
        FrameInterval(8, 8),
    ]

def test_object_data_pointers__attributes_one_annotation():
    object = build_object("person")

    frames = {
        0: build_frame(0,
            {
                object: [
                    build_annotation(
                        name="rgb_middle__bbox__person",
                        object=object,
                        attributes={
                            "text_attr": "some value",
                            "num_attr": 0,
                            "bool_attr": True,
                            "vec_attr": [0, 1],
                        }
                    ),
                ]
            }
        )
    }

    object_data_pointers = object.object_data_pointers(frames)

    assert object_data_pointers["rgb_middle__bbox__person"].attribute_pointers == {
        "text_attr": AttributeType.TEXT,
        "num_attr": AttributeType.NUM,
        "bool_attr": AttributeType.BOOLEAN,
        "vec_attr": AttributeType.VEC,
    }

def test_object_data_pointers__attributes_multiple_annotations_with_differing_attributes():
    object = build_object("person")

    frames = {
        0: build_frame(0,
            {
                object: [
                    build_annotation(
                        name="rgb_middle__bbox__person",
                        object=object,
                        attributes={
                            "text_attr": "some value",
                            "num_attr": 0,
                        }
                    ),
                    build_annotation(
                        name="rgb_middle__bbox__person",
                        object=object,
                        attributes={
                            "bool_attr": True,
                            "vec_attr": [0, 1],
                        }
                    ),
                ]
            }
        )
    }

    object_data_pointers = object.object_data_pointers(frames)

    assert object_data_pointers["rgb_middle__bbox__person"].attribute_pointers == {
        "text_attr": AttributeType.TEXT,
        "num_attr": AttributeType.NUM,
        "bool_attr": AttributeType.BOOLEAN,
        "vec_attr": AttributeType.VEC,
    }

def test_object_data_pointers__multiple_objects_of_differing_type():
    object_person = build_object("person")
    object_train = build_object("train")

    frames = {
        0: build_frame(0,
            {
                object_person: [
                    build_annotation("lidar__bbox__person", object_person),
                ]
            }
        ),
        1: build_frame(1,
            {
                object_train: [
                    build_annotation("lidar__bbox__train", object_train),
                ]
            }
        )
    }

    person_object_data_pointers = object_person.object_data_pointers(frames)
    assert len(person_object_data_pointers) == 1
    assert person_object_data_pointers["lidar__bbox__person"].frame_intervals == [
        FrameInterval(0, 0)
    ]

    train_object_data_pointers = object_train.object_data_pointers(frames)
    assert len(train_object_data_pointers) == 1
    assert train_object_data_pointers["lidar__bbox__train"].frame_intervals == [
        FrameInterval(1, 1)
    ]

def test_object_data_pointers__multiple_objects_of_same_type():
    object1 = build_object("person")
    object2 = build_object("person")

    frames = {
        0: build_frame(0,
            {
                object1: [
                    build_annotation("lidar__bbox__person", object1),
                ]
            }
        ),
        1: build_frame(1,
            {
                object2: [
                    build_annotation("lidar__bbox__person", object2),
                ]
            }
        )
    }

    object1_data_pointers = object1.object_data_pointers(frames)
    assert len(object1_data_pointers) == 1
    assert object1_data_pointers["lidar__bbox__person"].frame_intervals == [
        FrameInterval(0, 0)
    ]

    object2_data_pointers = object2.object_data_pointers(frames)
    assert len(object2_data_pointers) == 1
    assert object2_data_pointers["lidar__bbox__person"].frame_intervals == [
        FrameInterval(1, 1)
    ]

# == Helpers ==========================

def build_annotation(name: str, object: Object, attributes: dict={}) -> t.Union[Bbox, Cuboid]:
    sensor_uid, ann_type, object_type = tuple(name.split("__"))

    sensor = Sensor(sensor_uid)

    if ann_type == "bbox":
        return Bbox(
            uid=str(uuid4()),
            object=object,
            attributes=attributes,
            sensor=sensor,
            pos=Point2d(50, 100),
            size=Size2d(30, 30)
        )

    elif ann_type == "cuboid":
        return Cuboid(
            uid=str(uuid4()),
            object=object,
            attributes=attributes,
            sensor=sensor,
            pos=Point3d(50, 100, 20),
            size=Size3d(30, 30, 30),
            quat=Quaternion(0, 0, 0, 1),
        )

    else:
        raise ValueError()

def build_frame(uid: int, raw_object_data: t.Dict[Object, t.List[t.Union[Bbox, Cuboid]]]) -> Frame:
    annotations = {}
    for object, object_data in raw_object_data.items():
        for annotation in object_data:
            annotations[annotation.uid] = annotation

    return Frame(uid=uid, annotations=annotations)

def build_object(type: str) -> Object:
    return Object(
        uid=str(uuid4()),
        name=f"{type}_{str(random.randint(0, 9999)).zfill(4)}",
        type=type
    )

if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import sys
from decimal import Decimal
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

import raillabel.format.understand_ai as uai_format

# == Fixtures =========================

@pytest.fixture
def sensor_camera_uai_dict() -> dict:
    return {
        "type": "ir_middle",
        "uri": "A0001781_image/000_1632321843.100464760.png",
        "timestamp": "1632321843.100464760"
    }

@pytest.fixture
def sensor_camera_uai() -> dict:
    return uai_format.SensorReference(
        type="ir_middle",
        uri="A0001781_image/000_1632321843.100464760.png",
        timestamp=Decimal("1632321843.100464760"),
    )

@pytest.fixture
def sensor_camera_raillabel_dict() -> dict:
    return {
        "stream_properties": {
            "sync": {
                "timestamp": "1632321843.100464760"
            }
        },
        "uri": "000_1632321843.100464760.png"
    }


@pytest.fixture
def sensor_lidar_uai_dict() -> dict:
    return {
        "type": "LIDAR",
        "uri": "lidar_merged/000_1632321880.132833000.pcd",
        "timestamp": "1632321880.132833000"
    }

@pytest.fixture
def sensor_lidar_uai() -> dict:
    return uai_format.SensorReference(
        type="LIDAR",
        uri="lidar_merged/000_1632321880.132833000.pcd",
        timestamp=Decimal("1632321880.132833000"),
    )

@pytest.fixture
def sensor_lidar_raillabel_dict() -> dict:
    return {
        "stream_properties": {
            "sync": {
                "timestamp": "1632321880.132833000"
            }
        },
        "uri": "000_1632321880.132833000.pcd"
    }

# == Tests ============================

def test_fromdict():
    sensor_reference = uai_format.SensorReference.fromdict(
        {
            "type": "ir_middle",
            "uri": "A0001781_image/000_1632321843.100464760.png",
            "timestamp": "1632321843.100464760"
        }
    )

    assert sensor_reference.type == "ir_middle"
    assert sensor_reference.uri == "A0001781_image/000_1632321843.100464760.png"
    assert sensor_reference.timestamp == Decimal("1632321843.100464760")


def test_to_raillabel():
    sensor_reference = uai_format.SensorReference(
        type="ir_middle",
        uri="A0001781_image/000_1632321843.100464760.png",
        timestamp=Decimal("1632321843.100464760"),
    )

    assert sensor_reference.to_raillabel()[0] == "ir_middle"
    assert sensor_reference.to_raillabel()[1] == {
        "stream_properties": {
            "sync": {
                "timestamp": "1632321843.100464760"
            }
        },
        "uri": "000_1632321843.100464760.png"
    }

if __name__ == "__main__":
    import os
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import typing as t
import uuid
from dataclasses import dataclass
from decimal import Decimal

from ..._util._warning import _warning
from ._annotation import _Annotation
from ._translation import translate_class_id, translate_sensor_id
from .bounding_box_2d import BoundingBox2d
from .bounding_box_3d import BoundingBox3d
from .polygon_2d import Polygon2d
from .polyline_2d import Polyline2d
from .segmentation_3d import Segmentation3d


@dataclass
class Frame:
    """A container of dynamic, timewise, information.

    Parameters
    ----------
    id: int
        Numerical identifier of the frame. Must be unique within the scene.
    timestamp: decimal.Decimal
        Timestamp containing the Unix epoch time of the frame with up to nanosecond precision.
    annotations: dict
        Dictionary containing all annotations. The keys are the uids of the annotations and the
        values are objects of type BoundingBox2d, BoundingBox3d, Polygon2d, Polyline2d or
        Segementation3d.
    """

    id: int
    timestamp: Decimal
    bounding_box_2ds: t.Dict[str, BoundingBox2d]
    bounding_box_3ds: t.Dict[str, BoundingBox3d]
    polygon_2ds: t.Dict[str, Polygon2d]
    polyline_2ds: t.Dict[str, Polyline2d]
    segmentation_3ds: t.Dict[str, Segmentation3d]

    _annotation_uids: t.Set[str] = None

    @property
    def annotations(self) -> dict:
        """Return all annotations of this frame in one dict."""
        return {
            **self.bounding_box_2ds,
            **self.polyline_2ds,
            **self.polygon_2ds,
            **self.bounding_box_3ds,
            **self.segmentation_3ds,
        }

    @property
    def translated_objects(self) -> dict:
        """Return all objects in this frame and translate them.

        Returns
        -------
        dict
            Dictionary containing all objects. Keys are the object IDs and values are the
            translated class names.
        """
        return {
            str(a.object_id): translate_class_id(a.class_name) for a in self.annotations.values()
        }

    @property
    def translated_sensors(self) -> dict:
        """Return all sensors in this frame and translates them.

        Returns
        -------
        dict
            Dictionary containing all sensors. Keys are the translated sensor IDs and values are
            the SensorReference objects.
        """
        sensors_list = []

        for annotation in list(self.annotations.values()):
            sensors_list.append(annotation.sensor)
            sensors_list[-1].type = translate_sensor_id(sensors_list[-1].type)

        return {sensor.type: sensor for sensor in sensors_list}

    @classmethod
    def fromdict(cls, data_dict: dict) -> "Frame":
        """Generate a Frame from a dictionary in the UAI format.

        Parameters
        ----------
        data_dict: dict
            Understand.AI T4 format dictionary containing the data_dict.

        Returns
        -------
        Frame
            Converted frame.
        """

        cls._annotation_uids = set()

        return Frame(
            id=int(data_dict["frameId"]),
            timestamp=Decimal(data_dict["timestamp"]),
            bounding_box_2ds=cls._annotation_fromdict(
                data_dict["annotations"]["2D_BOUNDING_BOX"], BoundingBox2d
            ),
            bounding_box_3ds=cls._annotation_fromdict(
                data_dict["annotations"]["3D_BOUNDING_BOX"], BoundingBox3d
            ),
            polygon_2ds=cls._annotation_fromdict(data_dict["annotations"]["2D_POLYGON"], Polygon2d),
            polyline_2ds=cls._annotation_fromdict(
                data_dict["annotations"]["2D_POLYLINE"], Polyline2d
            ),
            segmentation_3ds=cls._annotation_fromdict(
                data_dict["annotations"]["3D_SEGMENTATION"], Segmentation3d
            ),
        )

    def to_raillabel(self) -> dict:
        """Generate a Frame from a dictionary in the UAI format.

        Parameters
        ----------
        data_dict: dict
            Understand.AI T4 format dictionary containing the data_dict.

        Returns
        -------
        Frame
            Converted frame.
        """
        return {
            "frame_properties": self._frame_properties_to_raillabel(),
            "objects": self._objects_to_raillabel(),
        }

    @classmethod
    def _annotation_fromdict(
        cls, data_dict: dict, annotation_class: t.Type[_Annotation]
    ) -> t.Dict[str, t.Type[_Annotation]]:

        annotations = {}
        for annotation_dict in data_dict:
            annotation_dict["id"] = cls._check_duplicate_annotation_uid(annotation_dict["id"])
            annotations[annotation_dict["id"]] = annotation_class.fromdict(annotation_dict)

        return {ann["id"]: annotation_class.fromdict(ann) for ann in data_dict}

    @classmethod
    def _check_duplicate_annotation_uid(cls, uid: str) -> str:

        if uid in cls._annotation_uids:
            _warning(
                f"Annotation uid {uid} is contained more than once. A new uid will be assigned."
            )
            return str(uuid.uuid4())

        cls._annotation_uids.add(uid)
        return uid

    def _frame_properties_to_raillabel(self) -> dict:

        streams_dict = {}
        for stream_id, stream in self.translated_sensors.items():
            streams_dict[stream_id] = {
                "stream_properties": {"sync": {"timestamp": str(stream.timestamp)}},
                "uri": stream.uri.split("/")[-1],
            }

        return {
            "timestamp": str(self.timestamp),
            "streams": {
                sensor.type: sensor.to_raillabel()[1] for sensor in self.translated_sensors.values()
            },
        }

    def _objects_to_raillabel(self) -> dict:
        object_data = {}

        for annotation in self.annotations.values():

            object_id = str(annotation.object_id)

            if object_id not in object_data:
                object_data[object_id] = {"object_data": {}}

            if annotation.OPENLABEL_ID not in object_data[object_id]["object_data"]:
                object_data[object_id]["object_data"][annotation.OPENLABEL_ID] = []

            object_data[object_id]["object_data"][annotation.OPENLABEL_ID].append(
                annotation.to_raillabel()[0]
            )

        return object_data



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from test_uai_attributes import attributes_raillabel_dict, attributes_uai, attributes_uai_dict
from test_uai_bounding_box_2d import (
    bounding_box_2d_raillabel_dict,
    bounding_box_2d_uai,
    bounding_box_2d_uai_dict,
)
from test_uai_bounding_box_3d import (
    bounding_box_3d_raillabel_dict,
    bounding_box_3d_uai,
    bounding_box_3d_uai_dict,
)
from test_uai_coordinate_system import (
    coordinate_system_camera_raillabel_dict,
    coordinate_system_camera_translated_uid,
    coordinate_system_camera_uai,
    coordinate_system_camera_uai_dict,
    coordinate_system_lidar_raillabel_dict,
    coordinate_system_lidar_translated_uid,
    coordinate_system_lidar_uai,
    coordinate_system_lidar_uai_dict,
    coordinate_system_radar_raillabel_dict,
    coordinate_system_radar_translated_uid,
    coordinate_system_radar_uai,
    coordinate_system_radar_uai_dict,
)
from test_uai_frame import frame_raillabel_dict, frame_uai, frame_uai_dict
from test_uai_metadata import metadata_raillabel_dict, metadata_uai, metadata_uai_dict
from test_uai_point_3d import point_3d_uai, point_3d_uai_dict, point_3d_vec
from test_uai_polygon_2d import polygon_2d_raillabel_dict, polygon_2d_uai, polygon_2d_uai_dict
from test_uai_polyline_2d import polyline_2d_raillabel_dict, polyline_2d_uai, polyline_2d_uai_dict
from test_uai_quaternion import quaternion_uai, quaternion_uai_dict, quaternion_vec
from test_uai_scene import scene_raillabel_dict, scene_uai, scene_uai_dict
from test_uai_segmentation_3d import (
    segmentation_3d_raillabel_dict,
    segmentation_3d_uai,
    segmentation_3d_uai_dict,
)
from test_uai_sensor_reference import (
    sensor_camera_raillabel_dict,
    sensor_camera_uai,
    sensor_camera_uai_dict,
    sensor_lidar_raillabel_dict,
    sensor_lidar_uai,
    sensor_lidar_uai_dict,
)
from test_uai_size_3d import size_3d_uai, size_3d_uai_dict, size_3d_vec



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import sys
from pathlib import Path
from uuid import UUID

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

import raillabel.format.understand_ai as uai_format
from raillabel.format.understand_ai._translation import translate_class_id

# == Fixtures =========================

@pytest.fixture
def polygon_2d_uai_dict(sensor_camera_uai_dict, attributes_uai_dict) -> dict:
    return {
        "id": "0f90cffa-2b6b-4e09-8fc2-527769a94e0a",
        "objectId": "58e7edd8-a7ee-4775-a837-e6dd375e8150",
        "className": "test_class",
        "geometry": {
            "points": [
                [127.71153737657284,  -0.3861000079676791],
                [127.4762636010818,  328.04436391207815],
                [115.77703250958459, 334.4789410124016],
                [115.01063176442402, 411.0810690770479],
            ]
        },
        "attributes": attributes_uai_dict,
        "sensor": sensor_camera_uai_dict
    }


@pytest.fixture
def polygon_2d_uai(attributes_uai, sensor_camera_uai) -> dict:
    return uai_format.Polygon2d(
        id=UUID("0f90cffa-2b6b-4e09-8fc2-527769a94e0a"),
        object_id=UUID("58e7edd8-a7ee-4775-a837-e6dd375e8150"),
        class_name="test_class",
        points=[
            (127.71153737657284,  -0.3861000079676791),
            (127.4762636010818,  328.04436391207815),
            (115.77703250958459, 334.4789410124016),
            (115.01063176442402, 411.0810690770479),
        ],
        attributes=attributes_uai,
        sensor=sensor_camera_uai,
    )

@pytest.fixture
def polygon_2d_raillabel_dict(attributes_raillabel_dict, coordinate_system_camera_translated_uid) -> dict:
    return {
        "name": "0f90cffa-2b6b-4e09-8fc2-527769a94e0a",
        "val": [
            127.71153737657284,  -0.3861000079676791,
            127.4762636010818,  328.04436391207815,
            115.77703250958459, 334.4789410124016,
            115.01063176442402, 411.0810690770479,
        ],
        "mode": "MODE_POLY2D_ABSOLUTE",
        "closed": True,
        "coordinate_system": coordinate_system_camera_translated_uid,
        "attributes": attributes_raillabel_dict,
    }


# == Tests ============================

def test_fromdict(
    attributes_uai_dict, attributes_uai,
    sensor_camera_uai_dict, sensor_camera_uai
):
    polygon_2d = uai_format.Polygon2d.fromdict(
        {
            "id": "0f90cffa-2b6b-4e09-8fc2-527769a94e0a",
            "objectId": "58e7edd8-a7ee-4775-a837-e6dd375e8150",
            "className": "test_class",
            "geometry": {
                "points": [
                    [127.71153737657284,  -0.3861000079676791],
                    [127.4762636010818,  328.04436391207815],
                    [115.77703250958459, 334.4789410124016],
                    [115.01063176442402, 411.0810690770479],
                ]
            },
            "attributes": attributes_uai_dict,
            "sensor": sensor_camera_uai_dict
        }
    )

    assert polygon_2d.id == UUID("0f90cffa-2b6b-4e09-8fc2-527769a94e0a")
    assert polygon_2d.object_id == UUID("58e7edd8-a7ee-4775-a837-e6dd375e8150")
    assert polygon_2d.class_name == "test_class"
    assert polygon_2d.points == [
        (127.71153737657284,  -0.3861000079676791),
        (127.4762636010818,  328.04436391207815),
        (115.77703250958459, 334.4789410124016),
        (115.01063176442402, 411.0810690770479),
    ]
    assert polygon_2d.attributes == attributes_uai
    assert polygon_2d.sensor == sensor_camera_uai


def test_to_raillabel(
    attributes_uai, attributes_raillabel_dict,
    sensor_camera_uai, sensor_camera_raillabel_dict, coordinate_system_camera_translated_uid,
):
    polygon_2d = uai_format.Polygon2d(
        id=UUID("0f90cffa-2b6b-4e09-8fc2-527769a94e0a"),
        object_id=UUID("58e7edd8-a7ee-4775-a837-e6dd375e8150"),
        class_name="test_class",
        points=[
            (127.71153737657284,  -0.3861000079676791),
            (127.4762636010818,  328.04436391207815),
            (115.77703250958459, 334.4789410124016),
            (115.01063176442402, 411.0810690770479),
        ],
        attributes=attributes_uai,
        sensor=sensor_camera_uai,
    )

    data_dict, object_id, translated_class_id, sensor_reference = polygon_2d.to_raillabel()

    assert data_dict == {
        "name": "0f90cffa-2b6b-4e09-8fc2-527769a94e0a",
        "val": [
            127.71153737657284,  -0.3861000079676791,
            127.4762636010818,  328.04436391207815,
            115.77703250958459, 334.4789410124016,
            115.01063176442402, 411.0810690770479,
        ],
        "mode": "MODE_POLY2D_ABSOLUTE",
        "coordinate_system": coordinate_system_camera_translated_uid,
        "closed": True,
        "attributes": attributes_raillabel_dict,
    }
    assert object_id == str(polygon_2d.object_id)
    assert translated_class_id == translate_class_id(polygon_2d.class_name)
    assert sensor_reference == sensor_camera_raillabel_dict

if __name__ == "__main__":
    import os
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent))

from raillabel._util._warning import _warning, _WarningsLogger


def test_issue_warning():
    with _WarningsLogger() as logger:
        _warning("lorem ipsum")

    assert logger.warnings == [
        "lorem ipsum"
    ]

def test_handover_exception():
    with pytest.raises(RuntimeError) as error:
        with _WarningsLogger() as logger:
            raise RuntimeError("weewoo something went wrong")

def test_clear_warnings():
    with _WarningsLogger() as logger1:
        _warning("lorem ipsum")

    with _WarningsLogger() as logger2:
        pass

    assert len(logger2.warnings) == 0


if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent))

import raillabel

# == Fixtures ============================

@pytest.fixture
def metadata() -> raillabel.format.Metadata:
    return raillabel.format.Metadata(schema_version="1.0.0")

# == Tests ============================

def test_simple_timespan(metadata):
    scene = raillabel.Scene(
        metadata=metadata,
        frames={
            0: raillabel.format.Frame(
                uid=0,
                timestamp=100
            ),
            1: raillabel.format.Frame(
                uid=1,
                timestamp=105
            ),
            2: raillabel.format.Frame(
                uid=0,
                timestamp=110
            ),
        }
    )

    assert raillabel.stats.generate_timespan(scene) == (100, 110)

def test_unordered_timspan(metadata):
    scene = raillabel.Scene(
        metadata=metadata,
        frames={
            0: raillabel.format.Frame(
                uid=0,
                timestamp=110
            ),
            1: raillabel.format.Frame(
                uid=1,
                timestamp=100
            ),
        }
    )

    assert raillabel.stats.generate_timespan(scene) == (100, 110)

def test_empty_timespan(metadata):
    scene = raillabel.Scene(
        metadata=metadata,
        frames={}
    )

    assert raillabel.stats.generate_timespan(scene) == (None, None)


if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os
import sys
from pathlib import Path

import pytest

sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent.parent))

from raillabel.format import Poly2d

# == Fixtures =========================

@pytest.fixture
def poly2d_dict(
    sensor_camera,
    attributes_multiple_types_dict,
    point2d_dict, point2d_another_dict
) -> dict:
    return {
        "uid": "d73b5988-767B-47ef-979c-022af60c6ab2",
        "name": "rgb_middle__poly2d__person",
        "val": point2d_dict + point2d_another_dict,
        "coordinate_system": sensor_camera.uid,
        "attributes": attributes_multiple_types_dict,
        "closed": True,
        "mode": "MODE_POLY2D_ABSOLUTE",
    }

@pytest.fixture
def poly2d(
    point2d, point2d_another,
    sensor_camera,
    attributes_multiple_types,
    object_person
) -> dict:
    return Poly2d(
        uid="d73b5988-767B-47ef-979c-022af60c6ab2",
        points=[point2d, point2d_another],
        object=object_person,
        sensor=sensor_camera,
        attributes=attributes_multiple_types,
        closed=True,
        mode="MODE_POLY2D_ABSOLUTE",
    )

# == Tests ============================

def test_fromdict(
    point2d, point2d_dict,
    point2d_another, point2d_another_dict,
    sensor_camera, sensors,
    object_person,
    attributes_multiple_types, attributes_multiple_types_dict,
):
    poly2d = Poly2d.fromdict(
        {
            "uid": "d73b5988-767B-47ef-979c-022af60c6ab2",
            "name": "rgb_middle__poly2d__person",
            "val": point2d_dict + point2d_another_dict,
            "coordinate_system": sensor_camera.uid,
            "attributes": attributes_multiple_types_dict,
            "closed": True,
            "mode": "MODE_POLY2D_ABSOLUTE",
        },
        sensors,
        object_person
    )

    assert poly2d.uid == "d73b5988-767B-47ef-979c-022af60c6ab2"
    assert poly2d.name == "rgb_middle__poly2d__person"
    assert poly2d.points == [point2d, point2d_another]
    assert poly2d.object == object_person
    assert poly2d.sensor == sensor_camera
    assert poly2d.attributes == attributes_multiple_types
    assert poly2d.closed == True
    assert poly2d.mode == "MODE_POLY2D_ABSOLUTE"


def test_asdict(
    point2d, point2d_dict,
    point2d_another, point2d_another_dict,
    sensor_camera, object_person,
    attributes_multiple_types, attributes_multiple_types_dict,
):
    poly2d = Poly2d(
        uid="d73b5988-767B-47ef-979c-022af60c6ab2",
        points=[point2d, point2d_another],
        object=object_person,
        sensor=sensor_camera,
        attributes=attributes_multiple_types,
        closed=True,
        mode="MODE_POLY2D_ABSOLUTE",
    )

    assert poly2d.asdict() == {
        "uid": "d73b5988-767B-47ef-979c-022af60c6ab2",
        "name": "rgb_middle__poly2d__person",
        "val": point2d_dict + point2d_another_dict,
        "coordinate_system": sensor_camera.uid,
        "attributes": attributes_multiple_types_dict,
        "closed": True,
        "mode": "MODE_POLY2D_ABSOLUTE",
    }


if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass

from ._object_annotation import _ObjectAnnotation
from .object import Object
from .point3d import Point3d
from .quaternion import Quaternion
from .size3d import Size3d


@dataclass
class Cuboid(_ObjectAnnotation):
    """3D bounding box.

    Parameters
    ----------
    uid: str
        This a string representing the unique universal identifier of the annotation.
    pos: raillabel.format.Point3d
        The center position of the cuboid in meters, where the x coordinate points ahead of the
        vehicle, y points to the left and z points upwards.
    quat: raillabel.format.Quaternion
        The rotation of the cuboid in quaternions.
    size: raillabel.format.Size3d
        The size of the cuboid in meters.
    object: raillabel.format.Object
        A reference to the object, this annotation belongs to.
    sensor: raillabel.format.Sensor, optional
        A reference to the sensor, this annotation is labeled in. Default is None.
    attributes: dict, optional
        Attributes of the annotation. Dict keys are the name str of the attribute, values are the
        attribute values. Default is {}.

    Properties (read-only)
    ----------------------
    name: str
        Name of the annotation used by the VCD player for indexing in the object data pointers.
    """

    pos: Point3d = None
    quat: Quaternion = None
    size: Size3d = None

    OPENLABEL_ID = "cuboid"
    _REQ_FIELDS = ["pos", "size", "quat"]

    @classmethod
    def fromdict(cls, data_dict: dict, sensors: dict, object: Object) -> "Cuboid":
        """Generate a Cuboid object from a dict.

        Parameters
        ----------
        data_dict: dict
            RailLabel format snippet containing the relevant data.
        sensors: dict
            Dictionary containing all sensors for the scene.
        object: raillabel.format.Object
            Object this annotation belongs to.

        Returns
        -------
        annotation: Cuboid
            Converted annotation.
        """

        return Cuboid(
            uid=str(data_dict["uid"]),
            pos=Point3d(
                x=data_dict["val"][0],
                y=data_dict["val"][1],
                z=data_dict["val"][2],
            ),
            quat=Quaternion(
                x=data_dict["val"][3],
                y=data_dict["val"][4],
                z=data_dict["val"][5],
                w=data_dict["val"][6],
            ),
            size=Size3d(
                x=data_dict["val"][7],
                y=data_dict["val"][8],
                z=data_dict["val"][9],
            ),
            object=object,
            sensor=cls._coordinate_system_fromdict(data_dict, sensors),
            attributes=cls._attributes_fromdict(data_dict),
        )

    def asdict(self) -> dict:
        """Export self as a dict compatible with the OpenLABEL schema.

        Returns
        -------
        dict_repr: dict
            Dict representation of this class instance.

        Raises
        ------
        ValueError
            if an attribute can not be converted to the type required by the OpenLabel schema.
        """

        dict_repr = self._annotation_required_fields_asdict()

        dict_repr["val"] = [
            float(self.pos.x),
            float(self.pos.y),
            float(self.pos.z),
            float(self.quat.x),
            float(self.quat.y),
            float(self.quat.z),
            float(self.quat.w),
            float(self.size.x),
            float(self.size.y),
            float(self.size.z),
        ]

        dict_repr.update(self._annotation_optional_fields_asdict())

        return dict_repr



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import pickle
import typing as t

from .. import format
from . import _filter_classes


def filter(scene: format.Scene, **kwargs) -> format.Scene:
    """Return a copy of the scene with the annotations filtered.

    Parameters
    ----------
    scene: raillabel.Scene
        Scene, which should be copied and filtered.
    include_object_types: str or list of str, optional
        List of class/type names that should be included in the filtered scene. If set, no
        other classes/types will be copied. Mutually exclusive with exclude_object_types.
    exclude_object_types: str or list of str, optional
        List of class/type names that should be excluded in the filtered scene. If set, all
        other classes/types will be copied. Mutually exclusive with include_object_types.
    include_annotation_types: str or list of str, optional
        List of annotation types (i.e. bboxs, cuboids, poly2ds, seg3ds) that should be included
        in the filtered scene. If set, no other annotation types will be copied. Mutually
        exclusive with exclude_annotation_types.
    exclude_annotation_types: str or list of str, optional
        List of annotation types (i.e. bboxs, cuboids, poly2ds, seg3ds) that should be excluded
        in the filtered scene. If set, all other annotation types will be copied. Mutually
        exclusive with include_annotation_types.
    include_annotation_ids: str or list of str, optional
        List of annotation UIDs that should be included in the filtered scene. If set, no other
        annotation UIDs will be copied. Mutually exclusive with exclude_annotation_ids.
    exclude_annotation_ids: str or list of str, optional
        List of annotation UIDs that should be excluded in the filtered scene. If set, all
        other annotation UIDs will be copied. Mutually exclusive with include_annotation_ids.
    include_object_ids: str or list of str, optional
        List of object UIDs that should be included in the filtered scene. If set, no other
        objects will be copied. Mutually exclusive with exclude_object_ids.
    exclude_object_ids: str or list of str, optional
        List of object UIDs that should be excluded in the filtered scene. If set, all other
        objects will be copied. Mutually exclusive with include_object_ids.
    include_sensors: str or list of str
        List of sensors that should be included in the filtered scene. If set, no other
        sensors will be copied. Mutually exclusive with exclude_sensors.
    exclude_sensors: str or list of str, optional
        List of sensors that should be excluded in the filtered scene. If set, all other
        sensors will be copied. Mutually exclusive with include_sensors.
    include_attributes: dict, optional
        Dict of attributes that should be included in the filtered scene. Dict keys are the
        attribute names, values are the specific values that should be included. If the
        value is set so None, all annotations with the attribute are included regardless of
        value. Mutually exclusive with exclude_attributes.
    exclude_attributes: dict, optional
        Dict of attributes that should be excluded in the filtered scene. Dict keys are the
        attribute names, values are the specific values that should be excluded. If the value
        is set so None, all annotations with the attribute are excluded regardless of value.
        Mutually exclusive with include_attributes.
    include_frames: int or list of int, optional
        List of frame UIDs that should be included in the filtered scene. If set, no other
        frames will be copied. Mutually exclusive with exclude_frames.
    exclude_frames: int or list of int, optional
        List of frame UIDs that should be excluded in the filtered scene. If set, all other
        frames will be copied. Mutually exclusive with include_frames.
    start_frame: int, optional
        Frame at which the filtered scene should start. Mutually exclusive with s
        tart_timestamp.
    end_frame: int, optional
        Frame at which the filtered scene should end (inclusive). Mutually exclusive with
        end_timestamp.
    start_timestamp: decimal.Decimal, optional
        Unix timestamp at which the filtered scene should start (inclusive). Mutually exclusive
        with start_frame.
    end_timestamp: decimal.Decimal, optional
        Unix timestamp at which the filtered scene should end (inclusive). Mutually exclusive
        with end_frame.

    Raises
    ------
    ValueError
        if two mutually exclusive parameters are set.
    TypeError
        if an unexpected keyword argument has been set.
    """

    filters_by_level = _collect_filter_classes(kwargs)
    filtered_scene, used_sensors, used_objects = _filter_scene(_copy(scene), filters_by_level)
    filtered_scene = _remove_unused(filtered_scene, used_sensors, used_objects)

    return filtered_scene


# --- Prepare filter classes


def _collect_filter_classes(kwargs) -> t.Tuple[t.List[t.Type], t.List[str]]:
    filters = []
    supported_kwargs = []
    for cls in _filter_classes.__dict__.values():
        if (
            isinstance(cls, type)
            and issubclass(cls, _filter_classes._FilterABC)
            and cls != _filter_classes._FilterABC
        ):
            filters.append(cls(kwargs))
            supported_kwargs.extend(cls.PARAMETERS)

    _check_for_unsupported_arg(kwargs, supported_kwargs)

    return _seperate_filters_by_level(filters)


def _check_for_unsupported_arg(kwargs: t.List[str], supported_kwargs: t.List[str]):
    for arg in kwargs:
        if arg not in supported_kwargs:
            raise TypeError(
                f"filter() got an unexpected keyword argument '{arg}'. Supported keyword "
                + f"arguments: {sorted(supported_kwargs)}"
            )


def _seperate_filters_by_level(filters: t.List[t.Type]) -> t.Dict[str, t.List[t.Type]]:
    all_filter_levels = [level for f in filters for level in f.LEVELS]

    filters_by_level = {level: [] for level in all_filter_levels}
    for level in filters_by_level:
        for filter_class in filters:
            if level in filter_class.LEVELS:
                filters_by_level[level].append(filter_class)

    return filters_by_level


# --- Filter scene


def _filter_scene(
    scene: format.Scene, filters_by_level: t.Dict[str, t.List[t.Type]]
) -> t.Tuple[format.Scene, t.Set[str], t.Set[str]]:

    used_sensors = set()
    used_objects = set()

    for frame_id, frame in list(scene.frames.items()):

        if not _passes_filters(frame, filters_by_level["frame"]):
            del scene.frames[frame_id]
            continue

        for frame_data_id, frame_data in list(frame.frame_data.items()):

            if _passes_filters(frame_data, filters_by_level["frame_data"]):
                used_sensors.add(frame_data.sensor.uid)

            else:
                del scene.frames[frame_id].frame_data[frame_data_id]

        for annotation_id, annotation in list(frame.annotations.items()):

            if _passes_filters(annotation, filters_by_level["annotation"]):
                used_objects.add(annotation.object.uid)
                used_sensors.add(annotation.sensor.uid)

            else:
                del scene.frames[frame_id].annotations[annotation_id]

    return scene, used_sensors, used_objects


# --- Remove unused


def _remove_unused(
    scene: format.Scene, used_sensors: t.Set[str], used_objects: t.Set[str]
) -> format.Scene:

    scene = _remove_unused_sensors(scene, used_sensors)
    scene = _remove_unused_objects(scene, used_objects)

    for frame_id in scene.frames:

        scene.frames[frame_id] = _remove_unused_sensor_references(
            scene.frames[frame_id], used_sensors
        )

    return scene


def _remove_unused_sensors(scene: format.Scene, used_sensors: t.Set[str]) -> format.Scene:
    for sensor_id in list(scene.sensors):
        if sensor_id not in used_sensors:
            del scene.sensors[sensor_id]

    return scene


def _remove_unused_objects(scene: format.Scene, used_objects: t.Set[str]) -> format.Scene:
    for object_id in list(scene.objects):
        if object_id not in used_objects:
            del scene.objects[object_id]

    return scene


def _remove_unused_sensor_references(frame: format.Frame, used_sensors: t.Set[str]) -> format.Frame:
    for sensor_id in list(frame.sensors):
        if sensor_id not in used_sensors:
            del frame.sensors[sensor_id]

    return frame


# --- Helper functions


def _passes_filters(data, filters):
    for f in filters:
        if not f.passes_filter(data):
            return False

    return True


def _copy(object):
    return pickle.loads(pickle.dumps(object, -1))



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import pytest

# == Fixtures =========================

@pytest.fixture
def attributes_uai_dict() -> dict:
    return {
        "isDummy": False,
        "carries": "nothing",
        "connectedTo": [],
        "pose": "upright"
    }

@pytest.fixture
def attributes_uai() -> dict:
    return {
        "isDummy": False,
        "carries": "nothing",
        "connectedTo": [],
        "pose": "upright"
    }

@pytest.fixture
def attributes_raillabel_dict() -> dict:
    return {
        "text": [
            {
                "name": "carries",
                "val": "nothing"
            },
            {
                "name": "pose",
                "val": "upright"
            }
        ],
        "boolean": [
            {
                "name": "isDummy",
                "val": False
            }
        ],
        "vec": [
            {
                "name": "connectedTo",
                "val": []
            }
        ]
    }



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import json
from dataclasses import dataclass
from pathlib import Path


@dataclass
class Metadata:
    """Container for metadata information about the scene itself.

    Parameters
    ----------
    clip_id: str
        Identifier of the scene for internal purposes.
    external_clip_id: str
        Identifier of the scene for external purposes.
    project_id: str
        Identifier of the annotation project.
    export_time: str
        Timestamp of the export in the format 'YYYY-MM-DD hh:mm UTC'.
    exporter_version: str
        Version of the Understand.AI-exporter.
    coordinate_system_3d: str
    coordinate_system_reference: str
    folder_name: str
        Directory with the exported reference data (e.g. images, point clouds).
    """

    clip_id: str
    external_clip_id: str
    project_id: str
    export_time: str
    exporter_version: str
    coordinate_system_3d: str
    coordinate_system_reference: str
    folder_name: str

    @classmethod
    def fromdict(cls, data_dict: dict) -> "Metadata":
        """Generate a Metadata from a dictionary in the UAI format.

        Parameters
        ----------
        data_dict: dict
            Understand.AI T4 format dictionary containing the data_dict.

        Returns
        -------
        metadata: Metadata
            Converted metadata.
        """

        return Metadata(
            clip_id=data_dict["clip_id"],
            external_clip_id=data_dict["external_clip_id"],
            project_id=data_dict["project_id"],
            export_time=data_dict["export_time"],
            exporter_version=data_dict["exporter_version"],
            coordinate_system_3d=data_dict["coordinate_system_3d"],
            coordinate_system_reference=data_dict["coordinate_system_reference"],
            folder_name=data_dict["folder_name"],
        )

    def to_raillabel(self) -> dict:
        """Convert to a raillabel compatible dict.

        Returns
        -------
        metadata: dict
            Converted metadata.
        """

        return {
            "name": self.external_clip_id,
            "schema_version": "1.0.0",
            "subschema_version": self._get_subschema_version(),
            "tagged_file": self.folder_name,
            "annotator": "understandAI GmbH",
        }

    def _get_subschema_version(self) -> str:
        RAILLABEL_SCHEMA_PATH = (
            Path(__file__).parent.parent.parent / "validate" / "schemas" / "raillabel_schema.json"
        )

        with RAILLABEL_SCHEMA_PATH.open() as schema_file:
            subschema_version = json.load(schema_file)["version"]

        return subschema_version
